# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import os
import shutil
import textwrap
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import Enum
from typing import Generator, Iterable, List, Optional, Tuple, Type, Union

from etl_entities.hwm import FileHWM, FileListHWM
from etl_entities.instance import AbsolutePath
from etl_entities.old_hwm import FileListHWM as OldFileListHWM
from etl_entities.source import RemoteFolder
from ordered_set import OrderedSet
from pydantic import Field, PrivateAttr, root_validator, validator

from onetl._internal import generate_temp_path
from onetl.base import BaseFileConnection, BaseFileFilter, BaseFileLimit
from onetl.base.path_protocol import PathProtocol
from onetl.file.file_downloader.options import FileDownloaderOptions
from onetl.file.file_downloader.result import DownloadResult
from onetl.file.file_set import FileSet
from onetl.file.filter.file_hwm import FileHWMFilter
from onetl.hooks import slot, support_hooks
from onetl.impl import (
    FailedRemoteFile,
    FileExistBehavior,
    FrozenModel,
    LocalPath,
    RemoteFile,
    RemotePath,
    path_repr,
)
from onetl.log import (
    entity_boundary_log,
    log_collection,
    log_hwm,
    log_lines,
    log_options,
    log_with_indent,
)
from onetl.strategy import StrategyManager
from onetl.strategy.batch_hwm_strategy import BatchHWMStrategy
from onetl.strategy.hwm_strategy import HWMStrategy

log = logging.getLogger(__name__)

# source, target, temp
DOWNLOAD_ITEMS_TYPE = OrderedSet[Tuple[RemotePath, LocalPath, Optional[LocalPath]]]


class FileDownloadStatus(Enum):
    SUCCESSFUL = 0
    FAILED = 1
    SKIPPED = 2
    MISSING = -1


@support_hooks
class FileDownloader(FrozenModel):
    """Allows you to download files from a remote source with specified file connection
    and parameters, and return an object with download result summary. |support_hooks|

    .. note::

        FileDownloader can return different results depending on :ref:`strategy`

    .. note::

        This class is used to download files **only** from remote directory to the local one.

        It does NOT support direct file transfer between filesystems, like ``FTP -> SFTP``.
        You should use FileDownloader + :ref:`file-uploader` to implement ``FTP -> local dir -> SFTP``.

    Parameters
    ----------
    connection : :obj:`onetl.connection.FileConnection`
        Class which contains File system connection properties. See :ref:`file-connections` section.

    local_path : :obj:`os.PathLike` or :obj:`str`
        Local path where you download files

    source_path : :obj:`os.PathLike` or :obj:`str`, optional, default: ``None``
        Remote path to download files from.

        Could be ``None``, but only if you pass absolute file paths directly to
        :obj:`~run` method

    temp_path : :obj:`os.PathLike` or :obj:`str`, optional, default: ``None``
        If set, this path will be used for downloading a file, and then renaming it to the target file path.
        If ``None`` is passed, files are downloaded directly to ``target_path``.

        .. warning::

            In case of production ETL pipelines, please set a value for ``temp_path`` (NOT ``None``).
            This allows to properly handle download interruption,
            without creating half-downloaded files in the target,
            because unlike file download, ``rename`` call is atomic.

        .. warning::

            In case of connections like SFTP or FTP, which can have multiple underlying filesystems,
            please pass to ``temp_path`` path on the SAME filesystem as ``target_path``.
            Otherwise instead of ``rename``, remote OS will move file between filesystems,
            which is NOT atomic operation.

    filters : list of :obj:`BaseFileFilter <onetl.base.base_file_filter.BaseFileFilter>`
        Return only files/directories matching these filters. See :ref:`file-filters`

    limits : list of :obj:`BaseFileLimit <onetl.base.base_file_limit.BaseFileLimit>`
        Apply limits to the list of files/directories, and stop if one of the limits is reached.
        See :ref:`file-limits`

    options : :obj:`~FileDownloader.Options`  | dict | None, default: ``None``
        File downloading options. See :obj:`~FileDownloader.Options`

    hwm : type[HWM] | None, default: ``None``

        HWM class to detect changes in incremental run. See :etl-entities:`File HWM <hwm/file/index.html>`

        .. warning ::
            Used only in :obj:`IncrementalStrategy <onetl.strategy.incremental_strategy.IncrementalStrategy>`.

    Examples
    --------
    Simple Downloader creation

    .. code:: python

        from onetl.connection import SFTP
        from onetl.file import FileDownloader

        sftp = SFTP(...)

        # create downloader
        downloader = FileDownloader(
            connection=sftp,
            source_path="/path/to/remote/source",
            local_path="/path/to/local",
        )

        # download files to "/path/to/local"
        downloader.run()

    Downloader with all parameters

    .. code:: python

        from onetl.connection import SFTP
        from onetl.file import FileDownloader
        from onetl.file.filter import Glob, ExcludeDir
        from onetl.file.limit import MaxFilesCount

        sftp = SFTP(...)

        # create downloader with a bunch of options
        downloader = FileDownloader(
            connection=sftp,
            source_path="/path/to/remote/source",
            local_path="/path/to/local",
            temp_path="/tmp",
            filters=[
                Glob("*.txt"),
                ExcludeDir("/path/to/remote/source/exclude_dir"),
            ],
            limits=[MaxFilesCount(100)],
            options=FileDownloader.Options(delete_source=True, if_exists="replace_file"),
        )

        # download files to "/path/to/local",
        # but only *.txt,
        # excluding files from "/path/to/remote/source/exclude_dir" directory
        # and stop before downloading 101 file
        downloader.run()

    Incremental download:

    .. code:: python

        from onetl.connection import SFTP
        from onetl.file import FileDownloader
        from onetl.strategy import IncrementalStrategy
        from etl_entities.hwm import FileListHWM

        sftp = SFTP(...)

        # create downloader
        downloader = FileDownloader(
            connection=sftp,
            source_path="/path/to/remote/source",
            local_path="/path/to/local",
            hwm=FileListHWM(
                name="my_unique_hwm_name", directory="/path/to/remote/source"
            ),  # mandatory for IncrementalStrategy
        )

        # download files to "/path/to/local", but only new ones
        with IncrementalStrategy():
            downloader.run()

    """

    Options = FileDownloaderOptions

    connection: BaseFileConnection

    local_path: LocalPath
    source_path: Optional[RemotePath] = None
    temp_path: Optional[LocalPath] = None

    filters: List[BaseFileFilter] = Field(default_factory=list, alias="filter")
    limits: List[BaseFileLimit] = Field(default_factory=list, alias="limit")

    hwm: Optional[FileHWM] = None
    hwm_type: Optional[Union[Type[OldFileListHWM], str]] = None

    options: FileDownloaderOptions = FileDownloaderOptions()

    _connection_checked: bool = PrivateAttr(default=False)

    @slot
    def run(self, files: Iterable[str | os.PathLike] | None = None) -> DownloadResult:  # noqa: WPS231
        """
        Method for downloading files from source to local directory. |support_hooks|

        .. note::

            This method can return different results depending on :ref:`strategy`

        Parameters
        ----------

        files : Iterable[str | os.PathLike] | None, default ``None``
            File list to download.

            If empty, download files from ``source_path`` to ``local_path``,
            applying ``filter``, ``limit`` and ``hwm`` to each one (if set).

            If not, download to ``local_path`` **all** input files, **ignoring**
            filters, limits and HWM.

        Returns
        -------
        downloaded_files : :obj:`DownloadResult <onetl.file.file_downloader.download_result.DownloadResult>`

            Download result object

        Raises
        ------
        :obj:`onetl.exception.DirectoryNotFoundError`

            ``source_path`` does not found

        NotADirectoryError

            ``source_path`` or ``local_path`` is not a directory

        Examples
        --------

        Download files from ``source_path`` to ``local_path``

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onetl.file import FileDownloader

            downloader = FileDownloader(source_path="/remote", local_path="/local", ...)
            downloaded_files = downloader.run()

            assert downloaded_files.successful == {
                LocalPath("/local/file1.txt"),
                LocalPath("/local/file2.txt"),
                LocalPath("/local/nested/path/file3.txt"),  # directory structure is preserved
            }
            assert downloaded_files.failed == {FailedRemoteFile("/remote/failed.file")}
            assert downloaded_files.skipped == {RemoteFile("/remote/already.exists")}
            assert downloaded_files.missing == {RemotePath("/remote/missing.file")}

        Download only certain files from ``source_path``

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onetl.file import FileDownloader

            downloader = FileDownloader(source_path="/remote", local_path="/local", ...)

            # paths could be relative or absolute, but all should be in "/remote"
            downloaded_files = downloader.run(
                [
                    "/remote/file1.txt",
                    "/remote/nested/path/file3.txt",
                    # excluding "/remote/file2.txt"
                ]
            )

            assert downloaded_files.successful == {
                LocalPath("/local/file1.txt"),
                LocalPath("/local/nested/path/file3.txt"),  # directory structure is preserved
            }
            assert not downloaded_files.failed
            assert not downloaded_files.skipped
            assert not downloaded_files.missing

        Download certain files from any folder

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onetl.file import FileDownloader

            downloader = FileDownloader(local_path="/local", ...)  # no source_path set

            # only absolute paths
            downloaded_files = downloader.run(
                [
                    "/remote/file1.txt",
                    "/any/nested/path/file2.txt",
                ]
            )

            assert downloaded_files.successful == {
                LocalPath("/local/file1.txt"),
                LocalPath("/local/file2.txt"),
                # directory structure is NOT preserved without source_path
            }
            assert not downloaded_files.failed
            assert not downloaded_files.skipped
            assert not downloaded_files.missing
        """

        entity_boundary_log(log, f"{self.__class__.__name__}.run() starts")

        if not self._connection_checked:
            self._log_parameters(files)

        self._check_strategy()

        if files is None and not self.source_path:
            raise ValueError("Neither file list nor `source_path` are passed")

        # Check everything
        if not self._connection_checked:
            self._check_local_path()
            self.connection.check()

            if self.source_path:
                self._check_source_path()

            self._connection_checked = True

        if files is None:
            log.debug("|%s| File list is not passed to `run` method", self.__class__.__name__)
            files = self.view_files()

        if not files:
            log.info("|%s| No files to download!", self.__class__.__name__)
            return DownloadResult()

        current_temp_dir: LocalPath | None = None
        if self.temp_path:
            current_temp_dir = generate_temp_path(self.temp_path)

        to_download = self._validate_files(files, current_temp_dir=current_temp_dir)

        # remove folder only after everything is checked
        if self.options.if_exists == FileExistBehavior.REPLACE_ENTIRE_DIRECTORY:
            if self.local_path.exists():
                shutil.rmtree(self.local_path)
            self.local_path.mkdir()

        if self.hwm:
            self._init_hwm(self.hwm)

        result = self._download_files(to_download)
        if current_temp_dir:
            self._remove_temp_dir(current_temp_dir)

        self._log_result(result)
        entity_boundary_log(log, f"{self.__class__.__name__}.run() ends", char="-")
        return result

    @slot
    def view_files(self) -> FileSet[RemoteFile]:
        """
        Get file list in the ``source_path``,
        after ``filter``, ``limit`` and ``hwm`` applied (if any). |support_hooks|

        .. note::

            This method can return different results depending on :ref:`strategy`

        Raises
        ------
        :obj:`onetl.exception.DirectoryNotFoundError`

            ``source_path`` does not found

        NotADirectoryError

            ``source_path`` is not a directory

        Returns
        -------
        FileSet[RemoteFile]
            Set of files in ``source_path``, which will be downloaded by :obj:`~run` method

        Examples
        --------

        View files

        .. code:: python

            from onetl.impl import RemoteFile
            from onetl.file import FileDownloader

            downloader = FileDownloader(source_path="/remote", ...)

            view_files = downloader.view_files()

            assert view_files == {
                RemoteFile("/remote/file1.txt"),
                RemoteFile("/remote/file3.txt"),
                RemoteFile("/remote/nested/file3.txt"),
            }
        """

        if not self.source_path:
            raise ValueError("Cannot call `.view_files()` without `source_path`")

        log.debug("|%s| Getting files list from path '%s'", self.connection.__class__.__name__, self.source_path)

        if not self._connection_checked:
            self._check_source_path()

        result = FileSet()

        filters = self.filters.copy()
        if self.hwm:
            filters.append(FileHWMFilter(hwm=self._init_hwm(self.hwm)))

        try:
            for root, _dirs, files in self.connection.walk(self.source_path, filters=filters, limits=self.limits):
                for file in files:
                    result.append(RemoteFile(path=root / file, stats=file.stats))

        except Exception as e:
            raise RuntimeError(
                f"Couldn't read directory tree from remote dir '{self.source_path}'",
            ) from e

        return result

    @validator("local_path", pre=True, always=True)
    def _resolve_local_path(cls, local_path):
        return LocalPath(local_path).resolve()

    @validator("source_path", pre=True, always=True)
    def _validate_source_path(cls, source_path):
        return RemotePath(source_path) if source_path else None

    @validator("temp_path", pre=True, always=True)
    def _validate_temp_path(cls, temp_path):
        return LocalPath(temp_path).resolve() if temp_path else None

    @root_validator(skip_on_failure=True)
    def _validate_hwm(cls, values):
        connection = values["connection"]
        source_path = values.get("source_path")
        hwm_type = values.get("hwm_type")
        hwm = values.get("hwm")

        if (hwm or hwm_type) and not source_path:
            raise ValueError("If `hwm` is passed, `source_path` must be specified")

        if hwm_type and (hwm_type == "file_list" or issubclass(hwm_type, OldFileListHWM)):
            remote_file_folder = RemoteFolder(name=source_path, instance=connection.instance_url)
            old_hwm = OldFileListHWM(source=remote_file_folder)
            warnings.warn(
                textwrap.dedent(
                    f"""
                    Passing "hwm_type" to FileDownloader class is deprecated since version 0.10.0,
                    and will be removed in v1.0.0.

                    Instead use:
                        hwm=FileListHWM(name={old_hwm.qualified_name!r})
                    """,
                ),
                UserWarning,
                stacklevel=2,
            )
            hwm = FileListHWM(
                name=old_hwm.qualified_name,
                directory=source_path,
            )

        if hwm and not hwm.entity:
            hwm = hwm.copy(update={"entity": AbsolutePath(source_path)})

        if hwm and hwm.entity != source_path:
            error_message = textwrap.dedent(
                f"""
                Passed `hwm.directory` is different from `source_path`.

                `hwm`:
                    {hwm!r}

                `source_path`:
                    {source_path!r}

                This is not allowed.
                """,
            )
            raise ValueError(error_message)

        values["hwm"] = hwm
        values["hwm_type"] = None
        return values

    @validator("filters", pre=True)
    def _validate_filters(cls, filters):
        if filters is None:
            warnings.warn(
                "filter=None is deprecated in v0.8.0, use filters=[] instead",
                UserWarning,
                stacklevel=3,
            )
            return []

        if not isinstance(filters, list):
            warnings.warn(
                "filter=... is deprecated in v0.8.0, use filters=[...] instead",
                UserWarning,
                stacklevel=3,
            )
            filters = [filters]

        return filters

    @validator("limits", pre=True)
    def _validate_limits(cls, limits):
        if limits is None:
            warnings.warn(
                "limit=None is deprecated in v0.8.0, use limits=[] instead",
                UserWarning,
                stacklevel=3,
            )
            return []

        if not isinstance(limits, list):
            warnings.warn(
                "limit=... is deprecated in v0.8.0, use limits=[...] instead",
                UserWarning,
                stacklevel=3,
            )
            limits = [limits]

        return limits

    def _check_strategy(self):
        strategy = StrategyManager.get_current()
        class_name = self.__class__.__name__
        strategy_name = strategy.__class__.__name__

        if self.hwm:
            if not isinstance(strategy, HWMStrategy):
                raise ValueError(f"{class_name}(hwm=...) cannot be used with {strategy_name}")

            offset = getattr(strategy, "offset", None)
            if offset is not None:
                raise ValueError(f"{class_name}(hwm=...) cannot be used with {strategy_name}(offset={offset}, ...)")

            if isinstance(strategy, BatchHWMStrategy):
                raise ValueError(f"{class_name}(hwm=...) cannot be used with {strategy_name}")

    def _init_hwm(self, hwm: FileHWM) -> FileHWM:
        strategy: HWMStrategy = StrategyManager.get_current()

        if not strategy.hwm:
            strategy.hwm = self.hwm
            strategy.fetch_hwm()
            return strategy.hwm

        if not isinstance(strategy.hwm, FileHWM) or strategy.hwm.name != hwm.name:
            # exception raised when inside one strategy >1 processes on the same table but with different hwm columns
            # are executed, example: test_postgres_strategy_incremental_hwm_set_twice
            error_message = textwrap.dedent(
                f"""
                Detected wrong {type(strategy).__name__} usage.

                Previous run:
                    {strategy.hwm!r}
                Current run:
                    {self.hwm!r}

                Probably you've executed code which looks like this:
                    with {strategy.__class__.__name__}(...):
                        FileDownloader(hwm=one_hwm, ...).run()
                        FileDownloader(hwm=another_hwm, ...).run()

                Please change it to:
                    with {strategy.__class__.__name__}(...):
                        FileDownloader(hwm=one_hwm, ...).run()

                    with {strategy.__class__.__name__}(...):
                        FileDownloader(hwm=another_hwm, ...).run()
                """,
            )
            raise ValueError(error_message)

        strategy.validate_hwm_attributes(hwm, strategy.hwm, origin=self.__class__.__name__)
        return strategy.hwm

    def _log_parameters(self, files: Iterable[str | os.PathLike] | None = None) -> None:
        log.info("|%s| -> |Local FS| Downloading files using parameters:", self.connection.__class__.__name__)
        log_with_indent(log, "source_path = %s", f"'{self.source_path}'" if self.source_path else "None")
        log_with_indent(log, "local_path = '%s'", self.local_path)
        log_with_indent(log, "temp_path = %s", f"'{self.temp_path}'" if self.temp_path else "None")
        log_collection(log, "filters", self.filters)
        log_collection(log, "limits", self.limits)
        if self.hwm:
            log_hwm(log, self.hwm)
        log_options(log, self.options.dict(by_alias=True))

        if self.options.delete_source:
            log.warning("|%s| SOURCE FILES WILL BE PERMANENTLY DELETED AFTER DOWNLOADING !!!", self.__class__.__name__)

        if self.options.if_exists == FileExistBehavior.REPLACE_ENTIRE_DIRECTORY:
            log.warning("|%s| LOCAL DIRECTORY WILL BE CLEANED UP BEFORE DOWNLOADING FILES !!!", self.__class__.__name__)

        if files and self.source_path:
            log.warning(
                "|%s| Passed both `source_path` and files list at the same time. Using explicit files list",
                self.__class__.__name__,
            )

    def _validate_files(  # noqa: WPS231
        self,
        remote_files: Iterable[os.PathLike | str],
        current_temp_dir: LocalPath | None,
    ) -> DOWNLOAD_ITEMS_TYPE:
        result: DOWNLOAD_ITEMS_TYPE = OrderedSet()

        for file in remote_files:
            remote_file_path = file if isinstance(file, PathProtocol) else RemotePath(file)
            remote_file = remote_file_path
            tmp_file: LocalPath | None = None

            if not self.source_path:
                # Download into a flat structure
                if not remote_file_path.is_absolute():
                    raise ValueError("Cannot pass relative file path with empty `source_path`")

                filename = remote_file_path.name
                local_file = self.local_path / filename
                if current_temp_dir:
                    tmp_file = current_temp_dir / filename  # noqa: WPS220
            else:
                # Download according to source folder structure
                if self.source_path in remote_file_path.parents:
                    # Make relative local path
                    local_file = self.local_path / remote_file_path.relative_to(self.source_path)
                    if current_temp_dir:
                        tmp_file = current_temp_dir / remote_file_path.relative_to(self.source_path)  # noqa: WPS220

                elif not remote_file_path.is_absolute():
                    # Passed path is already relative
                    local_file = self.local_path / remote_file_path
                    remote_file = self.source_path / remote_file_path
                    if current_temp_dir:
                        tmp_file = current_temp_dir / remote_file_path  # noqa: WPS220
                else:
                    # Wrong path (not relative path and source path not in the path to the file)
                    raise ValueError(f"File path '{remote_file}' does not match source_path '{self.source_path}'")

            if not isinstance(remote_file, PathProtocol) and self.connection.path_exists(remote_file):
                remote_file = self.connection.resolve_file(remote_file)

            result.add((remote_file, local_file, tmp_file))

        return result

    def _check_source_path(self):
        self.connection.resolve_dir(self.source_path)

    def _check_local_path(self):
        if self.local_path.exists() and not self.local_path.is_dir():
            raise NotADirectoryError(f"{path_repr(self.local_path)} is not a directory")

        self.local_path.mkdir(exist_ok=True, parents=True)

    def _download_files(  # noqa: WPS231
        self,
        to_download: DOWNLOAD_ITEMS_TYPE,
    ) -> DownloadResult:
        files = FileSet(item[0] for item in to_download)
        log.info("|%s| Files to be downloaded:", self.__class__.__name__)
        log_lines(log, str(files))
        log_with_indent(log, "")
        log.info("|%s| Starting the download process...", self.__class__.__name__)

        self._create_dirs(to_download)

        strategy = StrategyManager.get_current()
        result = DownloadResult()
        source_files: list[RemotePath] = []
        try:  # noqa: WPS501
            for status, source_file, target_file in self._bulk_download(to_download):
                if status == FileDownloadStatus.SUCCESSFUL:
                    result.successful.add(target_file)
                    source_files.append(source_file)
                elif status == FileDownloadStatus.FAILED:
                    result.failed.add(source_file)
                elif status == FileDownloadStatus.SKIPPED:
                    result.skipped.add(source_file)
                elif status == FileDownloadStatus.MISSING:
                    result.missing.add(source_file)
        finally:
            if self.hwm:
                # always update HWM in HWM store, even if downloader is interrupted
                strategy.update_hwm(source_files)
                strategy.save_hwm()
        return result

    def _create_dirs(
        self,
        to_download: DOWNLOAD_ITEMS_TYPE,
    ) -> None:
        """
        Create all parent paths before downloading files
        This is required to avoid errors then multiple threads create the same dir
        """
        parent_paths = OrderedSet()
        for _, target_file, tmp_file in to_download:
            parent_paths.add(target_file.parent)
            if tmp_file:
                parent_paths.add(tmp_file.parent)

        for parent_path in parent_paths:
            parent_path.mkdir(parents=True, exist_ok=True)

    def _bulk_download(
        self,
        to_download: DOWNLOAD_ITEMS_TYPE,
    ) -> Generator[tuple[FileDownloadStatus, RemotePath, LocalPath | None], None, None]:
        workers = self.options.workers
        files_count = len(to_download)

        real_workers = workers
        if files_count < workers:
            log.debug(
                "|%s| Asked for %d workers, but there are only %d files",
                self.__class__.__name__,
                workers,
                files_count,
            )
            real_workers = files_count

        if real_workers > 1:
            log.debug("|%s| Using ThreadPoolExecutor with %d workers", self.__class__.__name__, real_workers)
            with ThreadPoolExecutor(
                max_workers=real_workers,
                thread_name_prefix=self.__class__.__name__,
            ) as executor:
                futures = [
                    executor.submit(self._download_file, source_file, target_file, tmp_file)
                    for source_file, target_file, tmp_file in to_download
                ]
                yield from (future.result() for future in as_completed(futures))
        else:
            log.debug("|%s| Using plain old for-loop", self.__class__.__name__)
            yield from (
                self._download_file(source_file, target_file, tmp_file)
                for source_file, target_file, tmp_file in to_download
            )

    def _download_file(  # noqa: WPS231, WPS213
        self,
        source_file: RemotePath,
        local_file: LocalPath,
        tmp_file: LocalPath | None,
    ) -> tuple[FileDownloadStatus, RemotePath, LocalPath | None]:
        if tmp_file:
            log.info(
                "|%s| Downloading file '%s' to '%s' (via tmp '%s')",
                self.__class__.__name__,
                source_file,
                local_file,
                tmp_file,
            )
        else:
            log.info("|%s| Downloading file '%s' to '%s'", self.__class__.__name__, source_file, local_file)

        if not self.connection.path_exists(source_file):
            log.warning("|%s| Missing file '%s', skipping", self.__class__.__name__, source_file)
            return FileDownloadStatus.MISSING, source_file, None

        try:
            remote_file = self.connection.resolve_file(source_file)

            replace = False
            if local_file.exists():
                if self.options.if_exists == FileExistBehavior.ERROR:
                    raise FileExistsError(f"File {path_repr(local_file)} already exists")

                if self.options.if_exists == FileExistBehavior.IGNORE:
                    log.warning("|Local FS| File %s already exists, skipping", path_repr(local_file))
                    return FileDownloadStatus.SKIPPED, remote_file, None

                replace = True

            if tmp_file:
                # Files are loaded to temporary directory before moving them to target dir.
                # This prevents operations with partly downloaded files

                self.connection.download_file(remote_file, tmp_file, replace=replace)

                # remove existing file only after new file is downloaded
                # to avoid issues then there is no free space to download new file, but existing one is already gone
                if replace and local_file.exists():
                    log.warning("|Local FS| File %s already exists, overwriting", path_repr(local_file))
                    local_file.unlink()

                local_file.parent.mkdir(parents=True, exist_ok=True)
                shutil.move(tmp_file, local_file)
            else:
                # Direct download
                self.connection.download_file(remote_file, local_file, replace=replace)

            # Delete Remote
            if self.options.delete_source:
                self.connection.remove_file(remote_file)

            return FileDownloadStatus.SUCCESSFUL, remote_file, local_file

        except Exception as e:
            if log.isEnabledFor(logging.DEBUG):
                log.exception(
                    "|%s| Couldn't download file from source dir",
                    self.__class__.__name__,
                    exc_info=e,
                )
            else:
                log.exception(
                    "|%s| Couldn't download file from source dir: %s",
                    self.__class__.__name__,
                    e,
                    exc_info=False,
                )
            failed_file = FailedRemoteFile(
                path=remote_file.path,
                stats=remote_file.stats,
                exception=e,
            )
            return FileDownloadStatus.FAILED, failed_file, None

    def _remove_temp_dir(self, temp_dir: LocalPath) -> None:
        log.info("|Local FS| Removing temp directory '%s'", temp_dir)

        try:
            shutil.rmtree(temp_dir)
        except Exception:
            log.exception("|Local FS| Error while removing temp directory")

    def _log_result(self, result: DownloadResult) -> None:
        log_with_indent(log, "")
        log.info("|%s| Download result:", self.__class__.__name__)
        log_lines(log, str(result))



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
import os
import sys
from pathlib import Path, PurePosixPath, PureWindowsPath


class LocalPath(Path):
    def __new__(cls, *args, **kwargs):
        if cls is LocalPath:
            cls = LocalWindowsPath if os.name == "nt" else LocalPosixPath
        if sys.version_info < (3, 12):
            return cls._from_parts(args)
        else:
            return object.__new__(cls)  # noqa: WPS503


class LocalPosixPath(LocalPath, PurePosixPath):
    pass


class LocalWindowsPath(LocalPath, PureWindowsPath):
    pass



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar

from onetl._util.java import try_import_java_class
from onetl._util.scala import get_default_scala_version
from onetl._util.spark import get_spark_version
from onetl._util.version import Version
from onetl.exception import MISSING_JVM_CLASS_MSG
from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

READ_OPTIONS = frozenset(
    (
        "dataAddress",
        "treatEmptyValuesAsNulls",
        "setErrorCellsToFallbackValues",
        "usePlainNumberFormat",
        "inferSchema",
        "addColorColumns",
        "timestampFormat",
        "maxRowsInMemory",
        "maxByteArraySize",
        "tempFileThreshold",
        "excerptSize",
        "workbookPassword",
    ),
)

WRITE_OPTIONS = frozenset(
    (
        "dataAddress",
        "dateFormat",
        "timestampFormat",
    ),
)

log = logging.getLogger(__name__)


@support_hooks
class Excel(ReadWriteFileFormat):
    """
    Excel file format. |support_hooks|

    Based on `Spark Excel <https://github.com/crealytics/spark-excel>`_ file format.

    Supports reading/writing files with ``.xlsx`` (read/write) and ``.xls`` (read only) extensions.

    .. versionadded:: 0.9.4

    .. dropdown:: Version compatibility

        * Spark versions: 3.2.x - 3.5.x.

            .. warning::

                Not all combinations of Spark version and package version are supported.
                See `Maven index <https://mvnrepository.com/artifact/com.crealytics/spark-excel>`_
                and `official documentation <https://github.com/crealytics/spark-excel>`_.

        * Scala versions: 2.12 - 2.13
        * Java versions: 8 - 20

        See documentation from link above.

    .. note ::

        You can pass any option to the constructor, even if it is not mentioned in this documentation.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version. See link above.

    Examples
    --------

    Describe options how to read from/write to Excel file with specific options:

    .. code:: python

        from onetl.file.format import Excel
        from pyspark.sql import SparkSession

        # Create Spark session with Excel package loaded
        maven_packages = Excel.get_packages(spark_version="3.5.0")
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        excel = Excel(
            header=True,
            inferSchema=True,
        )

    """

    name: ClassVar[str] = "excel"

    header: bool = False

    class Config:
        known_options = READ_OPTIONS | WRITE_OPTIONS
        extra = "allow"

    @slot
    @classmethod
    def get_packages(
        cls,
        spark_version: str | Version,
        scala_version: str | Version | None = None,
        package_version: str | Version | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. |support_hooks|

        .. warning::

            Not all combinations of Spark version and package version are supported.
            See `Maven index <https://mvnrepository.com/artifact/com.crealytics/spark-excel>`_
            and `official documentation <https://github.com/crealytics/spark-excel>`_.

        Parameters
        ----------
        spark_version : str
            Spark version in format ``major.minor.patch``.

        scala_version : str, optional
            Scala version in format ``major.minor``.

            If ``None``, ``spark_version`` is used to determine Scala version.

        package_version : str, optional
            Package version in format ``major.minor.patch``. Default is ``0.20.3``.

            .. warning::

                Version ``0.14`` and below are not supported.

            .. note::

                It is not guaranteed that custom package versions are supported.
                Tests are performed only for default version.

        Examples
        --------

        .. code:: python

            from onetl.file.format import Excel

            Excel.get_packages(spark_version="3.5.0")
            Excel.get_packages(spark_version="3.5.0", scala_version="2.13")
            Excel.get_packages(
                spark_version="3.5.0",
                scala_version="2.13",
                package_version="0.20.3",
            )

        """

        if package_version:
            version = Version.parse(package_version)
            if version < (0, 15):
                # format="com.crealytics.spark.excel" does not support reading folder with files
                # format="excel" was added only in 0.14, but Maven package for 0.14 has different naming convention than recent versions.
                # So using 0.15 as the lowest supported version.
                raise ValueError(f"Package version should be at least 0.15, got {package_version}")
            log.warning("Passed custom package version %r, it is not guaranteed to be supported", package_version)
        else:
            version = Version(0, 20, 3)

        spark_ver = Version.parse(spark_version)
        if spark_ver < (3, 2):
            # Actually, Spark 2.4 is supported, but packages are built only for Scala 2.12
            # when default pyspark==2.4.1 is built with Scala 2.11.
            # See https://github.com/crealytics/spark-excel/issues/426
            raise ValueError(f"Spark version should be at least 3.2, got {spark_version}")

        scala_ver = Version.parse(scala_version) if scala_version else get_default_scala_version(spark_ver)
        if scala_ver.digits(2) < (2, 12):
            raise ValueError(f"Scala version should be at least 2.12, got {scala_ver}")

        return [f"com.crealytics:spark-excel_{scala_ver.digits(2)}:{spark_ver.digits(3)}_{version.digits(3)}"]

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        java_class = "com.crealytics.spark.excel.v2.ExcelDataSource"

        try:
            try_import_java_class(spark, java_class)
        except Exception as e:
            spark_version = get_spark_version(spark)
            msg = MISSING_JVM_CLASS_MSG.format(
                java_class=java_class,
                package_source=self.__class__.__name__,
                args=f"spark_version='{spark_version}'",
            )
            if log.isEnabledFor(logging.DEBUG):
                log.debug("Missing Java class", exc_info=e, stack_info=True)
            raise ValueError(msg) from e



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os
import re
import stat
import textwrap
from dataclasses import dataclass
from datetime import datetime
from pathlib import PurePath

from humanize import naturalsize

from onetl.base import ContainsException, PathProtocol, PathWithStatsProtocol

FILE_TYPE_DETECTORS = {
    stat.S_ISSOCK: "socket",
    stat.S_ISLNK: "link",
    stat.S_ISREG: "file",
    stat.S_ISBLK: "block",
    stat.S_ISDIR: "directory",
    stat.S_ISCHR: "char",
    stat.S_ISFIFO: "fifo",
}


@dataclass
class PathRepr:
    path: str
    kind: str | None = None
    mode: int | None = None
    size: int | None = None
    mtime: datetime | None = None
    user: str | int | None = None
    group: str | int | None = None
    exception: Exception | None = None

    @classmethod
    def from_path(cls, path) -> PathRepr:
        if isinstance(path, str):
            path = PurePath(path)

        result = cls(path=os.fspath(path))

        if isinstance(path, PathProtocol):
            result.kind = cls.detect_kind(path)

            if path.exists() and isinstance(path, PathWithStatsProtocol):
                details = path.stat()
                result.size = details.st_size
                result.mtime = datetime.fromtimestamp(details.st_mtime) if details.st_mtime else None
                result.mode = details.st_mode
                result.user = details.st_uid
                result.group = details.st_gid

        if isinstance(path, ContainsException):
            result.exception = path.exception

        return result

    @staticmethod
    def detect_kind(path: PathProtocol) -> str:
        if not path.exists():
            return "missing"

        if isinstance(path, PathWithStatsProtocol):
            mode = path.stat().st_mode
            if mode is not None:
                # try to detect mode based on stats
                for detector, file_kind in FILE_TYPE_DETECTORS.items():
                    if detector(mode):
                        return file_kind  # noqa: WPS220

        if path.is_dir():
            return "directory"

        if path.is_file():
            return "file"

        return "unknown"

    def repr_size(self) -> str:
        if self.size is not None and self.kind == "file":
            return f"size={naturalsize(self.size)!r}"

        return ""

    def repr_kind(self) -> str:
        if not self.kind:
            return ""

        return f"kind={self.kind!r}"

    def repr_mtime(self) -> str:
        if not self.mtime:
            return ""

        return f"mtime={self.mtime.isoformat()!r}"

    def repr_mode(self) -> str:
        if self.mode is None:
            return ""

        # first symbol is a file type, which is already stored in `kind` attribute
        return f"mode={stat.filemode(self.mode)[1:]!r}"

    def repr_user(self) -> str:
        if not self.user:
            return ""

        return f"uid={self.user!r}"

    def repr_group(self) -> str:
        if not self.group:
            return ""

        return f"gid={self.group!r}"

    def repr_exception(self) -> str:
        if not self.exception:
            return ""

        prefix = " " * 4
        exception = re.sub(r"(\\r)?\\n", os.linesep, repr(self.exception))
        exception_formatted = textwrap.indent(exception, prefix)
        return os.linesep + exception_formatted + os.linesep

    def info(
        self,
        with_kind: bool = True,
        with_size: bool = True,
        with_mode: bool = True,
        with_mtime: bool = True,
        with_owner: bool = True,
        with_exception: bool = True,
    ):
        result = repr(self.path)
        properties: list[str] = []
        if with_kind:
            properties.append(self.repr_kind())

        if with_size:
            properties.append(self.repr_size())

        if with_mode:
            properties.append(self.repr_mode())

        if with_mtime:
            properties.append(self.repr_mtime())

        if with_owner:
            properties.append(self.repr_user())
            properties.append(self.repr_group())

        properties = list(filter(None, properties))

        result_str = f"{result} ({', '.join(properties)})" if properties else result

        if with_exception:
            result_str += self.repr_exception()

        return result_str


def path_repr(
    path: os.PathLike | str,
    with_kind: bool = True,
    with_size: bool = True,
    with_mode: bool = True,
    with_mtime: bool = True,
    with_owner: bool = True,
    with_exception: bool = True,
) -> str:
    return PathRepr.from_path(path).info(
        with_kind=with_kind,
        with_size=with_size,
        with_mode=with_mode,
        with_mtime=with_mtime,
        with_owner=with_owner,
        with_exception=with_exception,
    )



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os
from typing import Iterable

from humanize import naturalsize
from pydantic import Field, validator

from onetl.base import PurePathProtocol
from onetl.exception import (
    EmptyFilesError,
    FailedFilesError,
    MissingFilesError,
    SkippedFilesError,
)
from onetl.file.file_set import FileSet
from onetl.impl import BaseModel

INDENT = " " * 4


class FileResult(BaseModel):
    """
    Result of some file manipulation process, e.g. download, upload, etc.

    Container for file paths, divided into certain categories:

    * :obj`successful`
    * :obj`failed`
    * :obj`skipped`
    * :obj`missing`
    """

    successful: FileSet[PurePathProtocol] = Field(default_factory=FileSet)
    "Successfully handled files"

    failed: FileSet[PurePathProtocol] = Field(default_factory=FileSet)
    "File paths which were handled with some failures"

    skipped: FileSet[PurePathProtocol] = Field(default_factory=FileSet)
    "File paths which were skipped because of some reason"

    missing: FileSet[PurePathProtocol] = Field(default_factory=FileSet)
    "Unknown paths which cannot be handled"

    @validator("successful", "failed", "skipped", "missing")
    def validate_container(cls, value: Iterable[PurePathProtocol]) -> FileSet[PurePathProtocol]:
        return FileSet(value)

    @property
    def successful_count(self) -> int:
        """
        Get number of successful files

        Examples
        --------

        .. code:: python

            from onetl.impl import LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                successful={LocalPath("/some/file"), LocalPath("/some/another.file")},
            )

            assert file_result.successful_count == 2
        """

        return len(self.successful)

    @property
    def failed_count(self) -> int:
        """
        Get number of failed files

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile
            from onet.file.file_result import FileResult

            file_result = FileResult(
                failed={RemoteFile("/some/file"), RemoteFile("/some/another.file")},
            )

            assert file_result.failed_count == 2
        """

        return len(self.failed)

    @property
    def skipped_count(self) -> int:
        """
        Get number of skipped files

        Examples
        --------

        .. code:: python

            from onetl.impl import LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                skipped={LocalPath("/some/file"), LocalPath("/some/another.file")},
            )

            assert file_result.skipped_count == 2
        """

        return len(self.skipped)

    @property
    def missing_count(self) -> int:
        """
        Get number of missing files

        Examples
        --------

        .. code:: python

            from onetl.impl import LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                missing={LocalPath("/some/file"), LocalPath("/some/another.file")},
            )

            assert file_result.missing_count == 2
        """

        return len(self.missing)

    @property
    def total_count(self) -> int:
        """
        Get total number of all files

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                successful={LocalPath("/local/file"), LocalPath("/local/another.file")},
                failed={RemoteFile("/remote/file"), RemoteFile("/remote/another.file")},
                skipped={LocalPath("/skipped/file")},
                missing={LocalPath("/missing/file")},
            )

            assert file_result.total_count == 6
        """

        return self.successful_count + self.failed_count + self.missing_count + self.skipped_count

    @property
    def successful_size(self) -> int:
        """
        Get size (in bytes) of successful files

        Examples
        --------

        .. code:: python

            from onetl.impl import LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                successful={LocalPath("/some/file"), LocalPath("/some/another.file")},
            )

            assert file_result.successful_size == 1_000_000  # in bytes
        """

        return self.successful.total_size

    @property
    def failed_size(self) -> int:
        """
        Get size (in bytes) of failed files

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile
            from onet.file.file_result import FileResult

            file_result = FileResult(
                failed={RemoteFile("/some/file"), RemoteFile("/some/another.file")},
            )

            assert file_result.failed_size == 1_000_000  # in bytes
        """

        return self.failed.total_size

    @property
    def skipped_size(self) -> int:
        """
        Get size (in bytes) of skipped files

        Examples
        --------

        .. code:: python

            from onetl.impl import LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                skipped={LocalPath("/some/file"), LocalPath("/some/another.file")},
            )

            assert file_result.skipped_size == 1_000_000  # in bytes
        """

        return self.skipped.total_size

    @property
    def total_size(self) -> int:
        """
        Get total size (in bytes) of all files

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                successful={LocalPath("/local/file"), LocalPath("/local/another.file")},
                failed={RemoteFile("/remote/file"), RemoteFile("/remote/another.file")},
                skipped={LocalPath("/skipped/file")},
                missing={LocalPath("/missing/file")},
            )

            assert file_result.total_size == 10_000_000  # in bytes
        """

        return self.successful_size + self.failed_size + self.skipped_size

    def raise_if_failed(self) -> None:
        """
        Raise exception if there are some files in ``failed`` attribute

        Raises
        ------
        FailedFilesError

            ``failed`` file set is not empty

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onet.file.file_result import FileResult

            files_with_exception = [
                FailedRemoteFile(
                    path="/remote/file1",
                    exception=NotAFileError("'/remote/file1' is not a file"),
                ),
                FailedRemoteFile(
                    path="/remote/file2",
                    exception=FileMissingError("'/remote/file2' does not exist"),
                ),
            ]

            file_result = FileResult(failed=files_with_exception)

            file_result.raise_if_failed()
            # will raise FailedFilesError('''
            #    Failed 2 files (10MB):
            #        '/remote/file1' (1 MB)
            #           NotAFileError("'/remote/file1' is not a file")
            #
            #        '/remote/file2' (9 MB)
            #           FileMissingError("'/remote/file2' does not exist")
            # ''')
        """

        if self.failed:
            raise FailedFilesError(self._failed_message)

    def raise_if_missing(self) -> None:
        """
        Raise exception if there are some files in ``missing`` attribute

        Raises
        ------
        MissingFilesError

            ``missing`` file set is not empty

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                missing={
                    LocalPath("/missing/file1"),
                    LocalPath("/missing/file2"),
                },
            )

            file_result.raise_if_missing()
            # will raise MissingFilesError('''
            #    Missing 2 files:
            #        '/missing/file1'
            #        '/missing/file2'
            # ''')
        """

        if self.missing:
            raise MissingFilesError(self._missing_message)

    def raise_if_skipped(self) -> None:
        """
        Raise exception if there are some files in ``skipped`` attribute

        Raises
        ------
        SkippedFilesError

            ``skipped`` file set is not empty

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                skipped={LocalPath("/skipped/file1"), LocalPath("/skipped/file2")},
            )

            file_result.raise_if_skipped()
            # will raise SkippedFilesError('''
            #    Skipped 2 files (15 kB):
            #        '/skipped/file1' (10kB)
            #        '/skipped/file2' (5 kB)
            # ''')
        """

        if self.skipped:
            raise SkippedFilesError(self._skipped_message)

    def raise_if_contains_zero_size(self) -> None:
        """
        Raise exception if ``successful`` attribute contains a file with zero size

        Raises
        ------
        ZeroFileSizeError

            ``successful`` file set contains a file with zero size

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult(
                successful={
                    LocalPath("/local/empty1.file"),
                    LocalPath("/local/empty2.file"),
                    LocalPath("/local/normal.file"),
                },
            )

            file_result.raise_if_contains_zero_size()
            # will raise ZeroFileSizeError('''
            #    2 files out of 3 have zero size:
            #        '/local/empty1.file'
            #        '/local/empty2.file'
            # ''')
        """

        self.successful.raise_if_contains_zero_size()

    @property
    def is_empty(self) -> bool:
        """
        Returns ``True`` if there are no files in ``successful``, ``failed`` and ``skipped`` attributes

        Examples
        --------

        .. code:: python

            from onetl.impl import LocalPath
            from onet.file.file_result import FileResult

            file_result1 = FileResult()
            assert file_result1.is_empty

            file_result2 = FileResult(
                successful={LocalPath("/local/file"), LocalPath("/local/another.file")},
            )
            assert not file_result2.is_empty
        """

        return not self.failed and not self.successful and not self.skipped

    def raise_if_empty(self) -> None:
        """
        Raise exception if there are no files in ``successful``, ``failed`` and ``skipped`` attributes

        Raises
        ------
        EmptyFilesError

            ``successful``, ``failed`` and ``skipped`` file sets are empty

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onet.file.file_result import FileResult

            file_result = FileResult()

            file_result.raise_if_empty()
            # will raise EmptyFilesError("There are no files in the result")
        """

        if self.is_empty:
            raise EmptyFilesError("There are no files in the result")

    @property
    def details(self) -> str:
        '''
        Return detailed information about files in the result object

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onet.file.file_result import FileResult

            file_result1 = FileResult(
                successful={LocalPath("/local/file"), LocalPath("/local/another.file")},
                failed={
                    FailedRemoteFile(
                        path="/remote/file1",
                        exception=NotAFileError("'/remote/file1' is not a file"),
                    ),
                    FailedRemoteFile(
                        path="/remote/file2",
                        exception=FileMissingError("'/remote/file2' does not exist"),
                    ),
                },
                skipped={LocalPath("/skipped/file1"), LocalPath("/skipped/file2")},
                missing={LocalPath("/missing/file1"), LocalPath("/missing/file2")},
            )

            details1 = """
                Total: 8 files (10.4 MB)

                Successful 2 files (30.7 kB):
                    '/successful1' (10.2 kB)
                    '/successful2' (20.5 kB)

                Failed 2 files (10MB):
                    '/remote/file1' (1 MB)
                        NotAFileError("'/remote/file1' is not a file")

                    '/remote/file2' (9 MB)
                        FileMissingError("'/remote/file2' does not exist")

                Skipped 2 files (15 kB):
                    '/skipped/file1' (10kB)
                    '/skipped/file2' (5 kB)

                Missing 2 files:
                    '/missing/file1'
                    '/missing/file2'
            """

            assert file_result1.details == details1

            file_result2 = FileResult()
            details2 = """
                No successful files

                No failed files

                No skipped files

                No missing files
            """

            assert file_result2.details == details2
        '''

        result = []

        if self.successful or self.failed or self.missing or self.skipped:
            result.append(self._total_summary)

        result.append(self._successful_message)
        result.append(self._failed_message)
        result.append(self._skipped_message)
        result.append(self._missing_message)

        return (os.linesep * 2).join(result)

    @property
    def summary(self) -> str:
        '''
        Return short summary about files in the result object

        Examples
        --------

        .. code:: python

            from onetl.impl import RemoteFile, LocalPath
            from onet.file.file_result import FileResult

            file_result1 = FileResult(
                successful={LocalPath("/local/file"), LocalPath("/local/another.file")},
                failed={RemoteFile("/remote/file"), RemoteFile("/remote/another.file")},
                skipped={LocalPath("/skipped/file")},
                missing={LocalPath("/missing/file")},
            )

            result = """
                Total: 8 files (10.4 MB)

                Successful: 2 files (30.7 kB)

                Failed: 2 files (10MB)

                Skipped: 2 files (15 kB)

                Missing: 2 files
            """

            assert file_result1.summary == result

            file_result2 = FileResult()
            assert file_result1.summary == "No files"
        '''
        return self._total_message

    def __str__(self):
        """Same as :obj:`onetl.file.file_result.FileResult.details`"""
        return self.details

    @property
    def _total_summary(self) -> str:
        if self.successful or self.failed or self.missing or self.skipped:
            file_number_str = f"{self.total_count} files" if self.total_count > 1 else "1 file"
            return f"Total: {file_number_str} (size='{naturalsize(self.total_size)}')"

        return "No files"

    @property
    def _successful_summary(self) -> str:
        if not self.successful:
            return "No successful files"

        return "Successful: " + self.successful.summary

    @property
    def _successful_message(self) -> str:
        if not self.successful:
            return self._successful_summary

        return "Successful " + self.successful.details

    @property
    def _failed_summary(self) -> str:
        if not self.failed:
            return "No failed files"

        return "Failed: " + self.failed.summary

    @property
    def _failed_message(self) -> str:
        if not self.failed:
            return self._failed_summary

        return "Failed " + self.failed.details

    @property
    def _skipped_summary(self) -> str:
        if not self.skipped:
            return "No skipped files"

        return "Skipped: " + self.skipped.summary

    @property
    def _skipped_message(self) -> str:
        if not self.skipped:
            return self._skipped_summary

        return "Skipped " + self.skipped.details

    @property
    def _missing_summary(self) -> str:
        if not self.missing:
            return "No missing files"

        return "Missing: " + self.missing.summary.replace(" (size='0 Bytes')", "")

    @property
    def _missing_message(self) -> str:
        if not self.missing:
            return self._missing_summary

        return "Missing " + self.missing.details.replace(" (size='0 Bytes')", "")

    @property
    def _total_message(self) -> str:
        result = [self._total_summary]

        if self.successful:
            result.append(self._successful_summary)

        if self.failed:
            result.append(self._failed_summary)

        if self.skipped:
            result.append(self._skipped_summary)

        if self.missing:
            result.append(self._missing_summary)

        return os.linesep.join(result)



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from functools import total_ordering
from string import ascii_lowercase
from typing import Iterator, NamedTuple, Sequence


@total_ordering
class Version(NamedTuple):
    """
    Version representation.

    Examples
    --------

    >>> Version(5, 6, 7)
    Version(major=5, minor=6, patch=7)
    >>> Version(5, 6)
    Version(major=5, minor=6, patch=None)
    >>> Version(5)
    Version(major=5, minor=None, patch=None)

    """

    major: int
    minor: int | None = None
    patch: int | None = None

    def __iter__(self) -> Iterator[int]:
        """
        Iterate over version components which are not ``None``.

        Examples
        --------

        >>> for part in Version(5, 6, 7):
        ...     print(part)
        5
        6
        7

        >>> for part in Version(5, 6):
        ...     print(part)
        5
        6

        >>> for part in Version(5):
        ...     print(part)
        5

        """
        yield self.major
        if self.minor is not None:
            yield self.minor
        if self.patch is not None:
            yield self.patch

    def __len__(self):
        """
        Get number of components set.

        Examples
        --------

        >>> assert len(Version(5, 6, 7)) == 3
        >>> assert len(Version(5, 6)) == 2
        >>> assert len(Version(5)) == 1

        """
        if self.patch is not None:
            return 3
        if self.minor is not None:
            return 2
        return 1

    def __str__(self):
        """
        Get version as string.

        Examples
        --------

        >>> assert str(Version(5, 6, 7)) == "5.6.7"
        >>> assert str(Version(5, 6)) == "5.6"
        >>> assert str(Version(5)) == "5"

        """
        return ".".join(map(str, self))

    def __eq__(self, other):
        """
        Compare versions.

        Examples
        --------

        >>> assert Version(5, 6, 7) == Version(5, 6, 7)

        >>> # Version could be replaced with tuple[int, ...]
        >>> assert Version(5, 6, 7) == (5, 6, 7)
        >>> assert Version(5, 6) == (5, 6)
        >>> assert Version(5) == (5,)

        """
        if not isinstance(other, tuple):
            return NotImplemented

        return tuple(self) == other

    def __gt__(self, other):
        """
        Compare versions.

        Examples
        --------

        >>> assert Version(5, 6, 7) > Version(5, 6, 6)
        >>> assert not Version(5, 6, 7) > Version(5, 6, 7)

        >>> # Version could be replaced with tuple[int, ...]
        >>> assert Version(5, 6, 7) > (5, 6)
        >>> assert not Version(5, 6, 7) > (5, 7)

        >>> assert Version(5, 6) > (5, 5)
        >>> assert not Version(5, 6) > (5, 6)
        >>> assert not Version(5, 6) > (5, 7)

        >>> assert Version(5, 6) > (5,)
        >>> assert not Version(5, 6) > (6,)

        >>> assert Version(5) > (4,)
        >>> assert not Version(5) > (5,)
        >>> assert not Version(5) > (6,)

        """
        if not isinstance(other, tuple):
            return NotImplemented

        return tuple(self) > other

    @classmethod
    def parse(cls, version: int | float | str | Sequence) -> Version:
        """
        Parse input as version object.

        Examples
        --------

        >>> assert Version.parse("5.6.7") == Version(5, 6, 7)
        >>> assert Version.parse("5.6") == Version(5, 6)
        >>> assert Version.parse("5") == Version(5)

        >>> assert Version.parse([5, 6, 7]) == Version(5, 6, 7)
        >>> assert Version.parse([5, 6]) == Version(5, 6)
        >>> assert Version.parse([5]) == Version(5)

        >>> assert Version.parse(5) == Version(5)
        >>> assert Version.parse(5.0) == Version(5, 0)

        """
        if isinstance(version, (int, float)):
            version = str(version)
        if isinstance(version, str):
            version = version.split(".")
        return cls(*map(int, version[:3]))

    def digits(self, items: int) -> Version:
        """
        Return version with exactly N components.

        Raises
        ------
        AssertionError
            There is not enough components

        Examples
        --------

        >>> assert Version(5, 6, 7).digits(3) == Version(5, 6, 7)
        >>> assert Version(5, 6, 7).digits(2) == Version(5, 6)
        >>> assert Version(5, 6, 7).digits(1) == Version(5)
        >>> Version(5, 6).digits(3)
        Traceback (most recent call last):
        AssertionError: Version '5.6' does not match format 'a.b.c'
        >>> Version(5).digits(2)
        Traceback (most recent call last):
        AssertionError: Version '5' does not match format 'a.b'
        """
        if len(self) < items:
            expected = ".".join(ascii_lowercase[:items])
            raise AssertionError(f"Version '{self}' does not match format '{expected}'")
        return self.parse(tuple(self)[:items])



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import getpass
import logging
import os
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from etl_entities.instance import Cluster, Host
from pydantic import Field, PrivateAttr, validator

from onetl.base import PurePathProtocol
from onetl.connection.file_df_connection.spark_file_df_connection import (
    SparkFileDFConnection,
)
from onetl.connection.file_df_connection.spark_hdfs.slots import SparkHDFSSlots
from onetl.hooks import slot, support_hooks
from onetl.impl import RemotePath

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

log = logging.getLogger(__name__)


@support_hooks
class SparkHDFS(SparkFileDFConnection):
    """
    Spark connection to HDFS. |support_hooks|

    Based on `Spark Generic File Data Source <https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html>`_.

    .. warning::

        To use Hive connector you should have PySpark installed (or injected to ``sys.path``)
        BEFORE creating the connector instance.

        You can install PySpark as follows:

        .. code:: bash

            pip install onetl[spark]  # latest PySpark version

            # or
            pip install onetl pyspark=3.5.0  # pass specific PySpark version

        See :ref:`install-spark` installation instruction for more details.

    .. note::

        Most of Hadoop instances use Kerberos authentication. In this case, you should call ``kinit``
        **BEFORE** starting Spark session to generate Kerberos ticket. See :ref:`install-kerberos`.

        In case of creating session with ``"spark.master": "yarn"``, you should also pass some additional options
        to Spark session, allowing executors to generate their own Kerberos tickets to access HDFS.
        See `Spark security documentation <https://spark.apache.org/docs/latest/security.html#kerberos>`_
        for more details.

    .. note::

        Supports only reading files as Spark DataFrame and writing DataFrame to files.

        Does NOT support file operations, like create, delete, rename, etc. For these operations,
        use :obj:`HDFS <onetl.connection.file_connection.hdfs.connection.HDFS>` connection.

    Parameters
    ----------
    cluster : str
        Cluster name.

        Used for:
            * HWM and lineage (as instance name for file paths)
            * Validation of ``host`` value,
                if latter is passed and if some hooks are bound to
                :obj:`Slots.get_cluster_namenodes <onetl.connection.file_df_connection.spark_hdfs.slots.SparkHDFSSlots.get_cluster_namenodes>`.

    host : str, optional
        Hadoop namenode host. For example: ``namenode1.domain.com``.

        Should be an active namenode (NOT standby).

        If value is not set, but there are some hooks bound to
        :obj:`Slots.get_cluster_namenodes <onetl.connection.file_df_connection.spark_hdfs.slots.SparkHDFSSlots.get_cluster_namenodes>`
        and
        :obj:`Slots.is_namenode_active <onetl.connection.file_df_connection.spark_hdfs.slots.SparkHDFSSlots.is_namenode_active>`,
        onETL will iterate over cluster namenodes to detect which one is active.

    ipc_port : int, default: ``8020``
        Port of Hadoop namenode (IPC protocol).

        If omitted, but there are some hooks bound to
        :obj:`Slots.get_ipc_port <onetl.connection.file_df_connection.spark_hdfs.slots.SparkHDFSSlots.get_ipc_port>`,
        onETL will try to detect port number for a specific ``cluster``.

    spark : :class:`pyspark.sql.SparkSession`
        Spark session

    Examples
    --------

    SparkHDFS connection initialization

    .. code:: python

        from onetl.connection import SparkHDFS
        from pyspark.sql import SparkSession

        # Create Spark session
        spark = SparkSession.builder.master("local").appName("spark-app-name").getOrCreate()

        # Create connection
        hdfs = SparkHDFS(
            host="namenode1.domain.com",
            cluster="rnd-dwh",
            spark=spark,
        ).check()

    SparkHDFS connection initialization with Kerberos support

    .. code:: python

        from onetl.connection import Hive
        from pyspark.sql import SparkSession

        # Create Spark session.
        # Use names "spark.yarn.access.hadoopFileSystems", "spark.yarn.principal"
        # and "spark.yarn.keytab" for Spark 2

        spark = (
            SparkSession.builder.appName("spark-app-name")
            .option(
                "spark.kerberos.access.hadoopFileSystems",
                "hdfs://namenode1.domain.com:8020",
            )
            .option("spark.kerberos.principal", "user")
            .option("spark.kerberos.keytab", "/path/to/keytab")
            .enableHiveSupport()
            .getOrCreate()
        )

        # Create connection
        hdfs = SparkHDFS(
            host="namenode1.domain.com",
            cluster="rnd-dwh",
            spark=spark,
        ).check()

    Automatically detect hostname for specific cluster
    (if some third-party plugin provides :ref:`spark-hdfs-slots` implementation):

    .. code:: python

        # Create Spark session
        ...

        # Create connection
        hdfs = SparkHDFS(cluster="rnd-dwh", spark=spark).check()
    """

    Slots = SparkHDFSSlots

    cluster: Cluster
    host: Optional[Host] = None
    ipc_port: int = Field(default=8020, alias="port")

    _active_host: Optional[Host] = PrivateAttr(default=None)

    @slot
    def path_from_string(self, path: os.PathLike | str) -> Path:
        return RemotePath(os.fspath(path))

    @property
    def instance_url(self):
        return self.cluster

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @slot
    def close(self):
        """
        Close all connections created to HDFS. |support_hooks|

        .. note::

            Connection can be used again after it was closed.

        Returns
        -------
        Connection itself

        Examples
        --------

        Close connection automatically:

        .. code:: python

            with connection:
                ...

        Close connection manually:

        .. code:: python

            connection.close()

        """
        log.debug("Reset FileSystem cache")
        with suppress(Exception):
            self._get_spark_fs().close()

        with suppress(Exception):
            self._active_host = None
        return self

    # Do not all __del__ with calling .close(), like other connections,
    # because this can influence dataframes created by this connection

    @slot
    @classmethod
    def get_current(cls, spark: SparkSession):
        """
        Create connection for current cluster. |support_hooks|

        Automatically sets up current cluster name as ``cluster``.

        .. note::

            Can be used only if there are a some hooks bound to
            :obj:`Slots.get_current_cluster <onetl.connection.file_df_connection.spark_hdfs.slots.SparkHDFSSlots.get_current_cluster>`.

        Parameters
        ----------
        spark : SparkSession

            See :obj:`~SparkHDFS` constructor documentation.

        Examples
        --------

        .. code:: python

            from onetl.connection import SparkHDFS

            # injecting current cluster name via hooks mechanism
            hdfs = SparkHDFS.get_current(spark=spark)
        """

        log.info("|%s| Detecting current cluster...", cls.__name__)
        current_cluster = cls.Slots.get_current_cluster()
        if not current_cluster:
            raise RuntimeError(
                f"{cls.__name__}.get_current() can be used only if there are "
                f"some hooks bound to {cls.__name__}.Slots.get_current_cluster",
            )

        log.info("|%s|   Got %r", cls.__name__, current_cluster)
        return cls(cluster=current_cluster, spark=spark)

    @validator("cluster")
    def _validate_cluster_name(cls, cluster):
        log.debug("|%s| Normalizing cluster %r name...", cls.__name__, cluster)
        validated_cluster = cls.Slots.normalize_cluster_name(cluster) or cluster
        if validated_cluster != cluster:
            log.debug("|%s|   Got %r", cls.__name__, validated_cluster)

        log.debug("|%s| Checking if cluster %r is a known cluster...", cls.__name__, validated_cluster)
        known_clusters = cls.Slots.get_known_clusters()
        if known_clusters and validated_cluster not in known_clusters:
            raise ValueError(
                f"Cluster {validated_cluster!r} is not in the known clusters list: {sorted(known_clusters)!r}",
            )

        return validated_cluster

    @validator("host")
    def _validate_host_name(cls, host, values):
        cluster = values.get("cluster")

        log.debug("|%s| Normalizing namenode %r host...", cls.__name__, host)
        namenode = cls.Slots.normalize_namenode_host(host, cluster) or host
        if namenode != host:
            log.debug("|%s|   Got %r", cls.__name__, namenode)

        log.debug("|%s| Checking if %r is a known namenode of cluster %r ...", cls.__name__, namenode, cluster)
        known_namenodes = cls.Slots.get_cluster_namenodes(cluster)
        if known_namenodes and namenode not in known_namenodes:
            raise ValueError(
                f"Namenode {namenode!r} is not in the known nodes list of cluster {cluster!r}: "
                f"{sorted(known_namenodes)!r}",
            )

        return namenode

    @validator("ipc_port", always=True)
    def _validate_port_number(cls, port, values):
        cluster = values.get("cluster")
        if cluster:
            log.debug("|%s| Getting IPC port of cluster %r ...", cls.__name__, cluster)
            result = cls.Slots.get_ipc_port(cluster) or port
            if result != port:
                log.debug("|%s|   Got %r", cls.__name__, result)
            return result

        return port

    def _get_active_namenode(self) -> str:
        class_name = self.__class__.__name__
        log.info("|%s| Detecting active namenode of cluster %r ...", class_name, self.cluster)

        namenodes = self.Slots.get_cluster_namenodes(self.cluster)
        if not namenodes:
            raise RuntimeError(f"Cannot get list of namenodes for a cluster {self.cluster!r}")

        nodes_len = len(namenodes)
        for i, namenode in enumerate(namenodes):
            log.debug("|%s|   Trying namenode %r (%d of %d) ...", class_name, namenode, i, nodes_len)
            if self.Slots.is_namenode_active(namenode, self.cluster):
                log.info("|%s|     Node %r is active!", class_name, namenode)
                return namenode
            log.debug("|%s|     Node %r is not active, skipping", class_name, namenode)

        raise RuntimeError(f"Cannot detect active namenode for cluster {self.cluster!r}")

    def _get_host(self) -> str:
        if not self.host:
            return self._get_active_namenode()

        # host is passed explicitly
        class_name = self.__class__.__name__
        log.info("|%s| Detecting if namenode %r of cluster %r is active...", class_name, self.host, self.cluster)

        is_active = self.Slots.is_namenode_active(self.host, self.cluster)
        if is_active:
            log.info("|%s|   Namenode %r is active!", class_name, self.host)
            return self.host

        if is_active is None:
            log.debug("|%s|   No hooks, skip validation", class_name)
            return self.host

        raise RuntimeError(f"Host {self.host!r} is not an active namenode of cluster {self.cluster!r}")

    def _convert_to_url(self, path: PurePathProtocol) -> str:
        # "hdfs://namenode:8020/absolute/path" if host is set
        if self._active_host:
            host = self._active_host
        else:
            host = self._get_host()
            # cache value to avoid getting active namenode for every path
            self._active_host = host
        return f"hdfs://{host}:{self.ipc_port}" + path.as_posix()

    def _get_default_path(self):
        return RemotePath("/user") / getpass.getuser()



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.file_df_connection.spark_s3.connection import SparkS3
from onetl.connection.file_df_connection.spark_s3.extra import SparkS3Extra



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Iterable, Optional

from ordered_set import OrderedSet
from pydantic import PrivateAttr, validator

from onetl._util.spark import try_import_pyspark
from onetl.base import BaseFileDFConnection, BaseReadableFileFormat, PurePathProtocol
from onetl.file.file_df_reader.options import FileDFReaderOptions
from onetl.file.file_set import FileSet
from onetl.hooks import slot, support_hooks
from onetl.impl import FrozenModel
from onetl.log import (
    entity_boundary_log,
    log_dataframe_schema,
    log_lines,
    log_options,
    log_with_indent,
)

if TYPE_CHECKING:
    from pyspark.sql import DataFrame
    from pyspark.sql.types import StructType

log = logging.getLogger(__name__)


@support_hooks
class FileDFReader(FrozenModel):
    """Allows you to read files from a source path with specified file connection
    and parameters, and return a Spark DataFrame. |support_hooks|

    .. warning::

        This class does **not** support read strategies.

    Parameters
    ----------
    connection : :obj:`BaseFileDFConnection <onetl.base.base_file_df_connection.BaseFileDFConnection>`
        File DataFrame connection. See :ref:`file-df-connections` section.

    format : :obj:`BaseReadableFileFormat <onetl.base.base_file_format.BaseReadableFileFormat>`
        File format to read.

    source_path : os.PathLike or str, optional, default: ``None``
        Directory path to read data from.

        Could be ``None``, but only if you pass file paths directly to
        :obj:`~run` method

    df_schema : :obj:`pyspark.sql.types.StructType`, optional, default: ``None``
        Spark DataFrame schema.

    options : :obj:`FileDFReaderOptions <onetl.file.file_df_reader.options.FileDFReaderOptions>`, optional
        Common reading options.

    Examples
    --------
    Create reader to parse CSV files in local filesystem:

    .. code:: python

        from onetl.connection import SparkLocalFS
        from onetl.file import FileDFReader
        from onetl.file.format import CSV

        local_fs = SparkLocalFS(spark=spark)

        reader = FileDFReader(
            connection=local_fs,
            format=CSV(delimiter=","),
            source_path="/path/to/directory",
        )

    All supported options

    .. code:: python

        from onetl.connection import SparkLocalFS
        from onetl.file import FileDFReader
        from onetl.file.format import CSV

        csv = CSV(delimiter=",")
        local_fs = SparkLocalFS(spark=spark)

        reader = FileDFReader(
            connection=local_fs,
            format=csv,
            source_path="/path/to/directory",
            options=FileDFReader.Options(recursive=False),
        )
    """

    Options = FileDFReaderOptions

    connection: BaseFileDFConnection
    format: BaseReadableFileFormat
    source_path: Optional[PurePathProtocol] = None
    df_schema: Optional[StructType] = None
    options: FileDFReaderOptions = FileDFReaderOptions()

    _connection_checked: bool = PrivateAttr(default=False)

    @slot
    def run(self, files: Iterable[str | os.PathLike] | None = None) -> DataFrame:
        """
        Method for reading files as DataFrame. |support_hooks|

        Parameters
        ----------

        files : Iterator[str | os.PathLike] | None, default ``None``
            File list to read.

            If empty, read files from ``source_path``.

        Returns
        -------
        df : :obj:`pyspark.sql.DataFrame`

            Spark DataFrame

        Examples
        --------

        Read CSV files from directory ``/path``:

        .. code:: python

            from onetl.connection import SparkLocalFS
            from onetl.file import FileDFReader
            from onetl.file.format import CSV

            csv = CSV(delimiter=",")
            local_fs = SparkLocalFS(spark=spark)

            reader = FileDFReader(
                connection=local_fs,
                format=csv,
                source_path="/path",
            )
            df = reader.run()

        Read some CSV files using file paths:

        .. code:: python

            from onetl.connection import SparkLocalFS
            from onetl.file import FileDFReader
            from onetl.file.format import CSV

            csv = CSV(delimiter=",")
            local_fs = SparkLocalFS(spark=spark)

            reader = FileDFReader(
                connection=local_fs,
                format=csv,
            )

            df = reader.run(
                [
                    "/path/file1.csv",
                    "/path/nested/file2.csv",
                ]
            )

        Read only specific CSV files in directory:

        .. code:: python

            from onetl.connection import SparkLocalFS
            from onetl.file import FileDFReader
            from onetl.file.format import CSV

            csv = CSV(delimiter=",")
            local_fs = SparkLocalFS(spark=spark)

            reader = FileDFReader(
                connection=local_fs,
                format=csv,
                source_path="/path",
            )

            df = reader.run(
                [
                    # file paths could be relative
                    "/path/file1.csv",
                    "/path/nested/file2.csv",
                ]
            )
        """

        entity_boundary_log(log, msg=f"{self.__class__.__name__}.run() starts")

        if files is None and not self.source_path:
            raise ValueError("Neither file list nor `source_path` are passed")

        if not self._connection_checked:
            self._log_parameters(files)

        paths: FileSet[PurePathProtocol] = FileSet()
        if files is not None:
            paths = FileSet(self._validate_files(files))
        elif self.source_path:
            paths = FileSet([self.source_path])

        if not self._connection_checked:
            self.connection.check()
            log_with_indent(log, "")
            self._connection_checked = True

        df = self._read_files(paths)
        entity_boundary_log(log, msg=f"{self.__class__.__name__}.run() ends", char="-")
        return df

    def _read_files(self, paths: FileSet[PurePathProtocol]) -> DataFrame:
        log.info("|%s| Paths to be read:", self.__class__.__name__)
        log_lines(log, str(paths))
        log_with_indent(log, "")

        return self.connection.read_files_as_df(
            root=self.source_path,
            paths=list(paths),
            format=self.format,
            df_schema=self.df_schema,
            options=self.options,
        )

    def _log_parameters(self, files: Iterable[str | os.PathLike] | None = None) -> None:
        log.info("|%s| -> |Spark| Reading files using parameters:", self.connection.__class__.__name__)
        log_with_indent(log, "source_path = %s", f"'{self.source_path}'" if self.source_path else "None")
        log_with_indent(log, "format = %r", self.format)

        if self.df_schema:
            empty_df = self.connection.spark.createDataFrame([], self.df_schema)  # type: ignore[attr-defined]
            log_dataframe_schema(log, empty_df)

        options_dict = self.options.dict(exclude_none=True)
        log_options(log, options_dict)

        if files is not None and self.source_path:
            log.warning(
                "|%s| Passed both `source_path` and files list at the same time. Using explicit files list",
                self.__class__.__name__,
            )

    @validator("source_path", pre=True)
    def _validate_source_path(cls, source_path, values):
        connection: BaseFileDFConnection = values["connection"]
        if source_path is None:
            return None
        return connection.path_from_string(source_path)

    @validator("format")
    def _validate_format(cls, format, values):  # noqa: WPS125
        connection: BaseFileDFConnection = values["connection"]
        connection.check_if_format_supported(format)
        return format

    @validator("options")
    def _validate_options(cls, value):
        return cls.Options.parse(value)

    def _validate_files(  # noqa: WPS231
        self,
        files: Iterable[os.PathLike | str],
    ) -> OrderedSet[PurePathProtocol]:
        result: OrderedSet[PurePathProtocol] = OrderedSet()

        for file in files:
            file_path = file if isinstance(file, PurePathProtocol) else self.connection.path_from_string(file)

            if not self.source_path:
                if not file_path.is_absolute():
                    raise ValueError("Cannot pass relative file path with empty `source_path`")
            elif file_path.is_absolute() and self.source_path not in file_path.parents:
                raise ValueError(f"File path '{file_path}' does not match source_path '{self.source_path}'")
            elif not file_path.is_absolute():
                # Make file path absolute
                file_path = self.source_path / file

            result.add(file_path)

        return result

    @classmethod
    def _forward_refs(cls) -> dict[str, type]:
        try_import_pyspark()
        from pyspark.sql.types import StructType  # noqa: WPS442

        # avoid importing pyspark unless user called the constructor,
        # as we allow user to use `Connection.get_packages()` for creating Spark session
        refs = super()._forward_refs()
        refs["StructType"] = StructType
        return refs



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import glob
import os
import re
import textwrap
import warnings
from typing import List, Optional, Union

from pydantic import Field, root_validator, validator

from onetl.base import BaseFileFilter, PathProtocol
from onetl.impl import FrozenModel, RemotePath


class FileFilter(BaseFileFilter, FrozenModel):
    r"""Filter files or directories by their path.

    .. deprecated:: 0.8.0

        Use :obj:`Glob <onetl.file.filter.glob.Glob>`, :obj:`Regexp <onetl.file.filter.regexp.Regexp>`
        or :obj:`ExcludeDir <onetl.file.filter.exclude_dir.ExcludeDir>` instead.

    Parameters
    ----------

    glob : str | None, default ``None``

        Pattern (e.g. ``*.csv``) for which any **file** (only file) path should match

        .. warning::

            Mutually exclusive with ``regexp``

    regexp : str | re.Pattern | None, default ``None``

        Regular expression (e.g. ``\d+\.csv``) for which any **file** (only file) path should match.

        If input is a string, regular expression will be compiles using ``re.IGNORECASE`` and ``re.DOTALL`` flags

        .. warning::

            Mutually exclusive with ``glob``

    exclude_dirs : list[os.PathLike | str], default ``[]``

        List of directories which should not be a part of a file or directory path


    Examples
    --------

    Create exclude_dir filter:

    .. code:: python

        from onetl.core import FileFilter

        file_filter = FileFilter(exclude_dirs=["/export/news_parse/exclude_dir"])

    Create glob filter:

    .. code:: python

        from onetl.core import FileFilter

        file_filter = FileFilter(glob="*.csv")

    Create regexp filter:

    .. code:: python

        from onetl.core import FileFilter

        file_filter = FileFilter(regexp=r"\d+\.csv")

        # or

        import re

        file_filter = FileFilter(regexp=re.compile("\d+\.csv"))

    Not allowed:

    .. code:: python

        from onetl.core import FileFilter

        FileFilter()  # will raise ValueError, at least one argument should be passed
    """

    class Config:
        arbitrary_types_allowed = True

    glob: Optional[str] = None
    regexp: Optional[re.Pattern] = None
    exclude_dirs: List[RemotePath] = Field(default_factory=list)

    @validator("glob", pre=True)
    def check_glob(cls, value: str) -> str:
        if not glob.has_magic(value):
            raise ValueError("Invalid glob")

        return value

    @validator("regexp", pre=True)
    def check_regexp(cls, value: Union[re.Pattern, str]) -> re.Pattern:
        if isinstance(value, str):
            return re.compile(value, re.IGNORECASE | re.DOTALL)

        return value

    @validator("exclude_dirs", each_item=True, pre=True)
    def check_exclude_dir(cls, value: Union[str, os.PathLike]) -> RemotePath:
        return RemotePath(value)

    @root_validator
    def disallow_empty_fields(cls, value: dict) -> dict:
        if value.get("glob") is None and value.get("regexp") is None and not value.get("exclude_dirs"):
            raise ValueError("One of the following fields must be set: `glob`, `regexp`, `exclude_dirs`")

        return value

    @root_validator
    def disallow_both_glob_and_regexp(cls, value: dict) -> dict:
        if value.get("glob") and value.get("regexp"):
            raise ValueError("Only one of `glob`, `regexp` fields can passed, not both")

        return value

    @root_validator
    def log_deprecated(cls, value: dict) -> dict:
        imports = []
        old_filters = []
        new_filters = []
        glob = value.get("glob")  # noqa: WPS442
        if glob is not None:
            imports.append("Glob")
            old_filters.append(f"glob={glob!r}")
            new_filters.append(f"Glob({glob!r})")

        regexp = value.get("regexp")
        if regexp is not None:
            imports.append("Regexp")
            old_filters.append(f"regexp={regexp.pattern!r}")
            new_filters.append(f"Regexp({regexp.pattern!r})")

        exclude_dirs = value.get("exclude_dirs")
        if exclude_dirs:
            imports.append("ExcludeDir")
            exclude_dirs_str = [repr(os.fspath(exclude_dir)) for exclude_dir in exclude_dirs]
            old_filters.append(f"exclude_dirs=[{', '.join(exclude_dirs_str)}]")
            new_filters.extend(f"ExcludeDir({item})" for item in exclude_dirs_str)

        if not imports:
            return value

        message = f"""
            Using FileFilter is deprecated since v0.8.0 and will be removed in v1.0.0.

            Please replace:
                from onetl.core import FileFilter

                filter=FileFilter({', '.join(old_filters)})

            With:
                from onetl.file.filter import {', '.join(imports)}

                filters=[{', '.join(new_filters)}]
        """

        warnings.warn(
            textwrap.dedent(message).strip(),
            category=UserWarning,
            stacklevel=3,  # 1 is current method, 2 is BaseModel internals, 3 is user code
        )
        return value

    def match(self, path: PathProtocol) -> bool:
        """False means it does not match the template by which you want to receive files"""

        if self.exclude_dirs:
            for exclude_dir in self.exclude_dirs:
                if path.is_dir() and exclude_dir == path:
                    return False
                if exclude_dir in path.parents:
                    return False

        if self.glob and path.is_file():
            return path.match(self.glob)

        if self.regexp and path.is_file():
            return self.regexp.search(os.fspath(path)) is not None

        return True



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os
from dataclasses import dataclass

from onetl.base import PathStatProtocol
from onetl.impl.path_container import PathContainer
from onetl.impl.remote_directory import RemoteDirectory
from onetl.impl.remote_path import RemotePath


@dataclass(eq=False, frozen=True)
class RemoteFile(PathContainer[RemotePath]):
    """
    Representation of existing remote file with stat
    """

    stats: PathStatProtocol

    def __post_init__(self):
        # frozen=True does not allow to change any field in __post_init__, small hack here
        object.__setattr__(self, "path", RemotePath(self.path))  # noqa: WPS609

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({os.fspath(self.path)!r})"

    def is_dir(self) -> bool:
        return False

    def is_file(self) -> bool:
        return True

    def exists(self) -> bool:
        return True

    def stat(self) -> PathStatProtocol:
        return self.stats

    @property
    def parent(self) -> RemoteDirectory:
        return RemoteDirectory(self.path.parent)

    @property
    def parents(self) -> list[RemoteDirectory]:
        return [RemoteDirectory(parent) for parent in self.path.parents]


@dataclass(eq=False, frozen=True)
class FailedRemoteFile(RemoteFile):
    """
    Representation of existing remote file with stat and attached exception object
    """

    exception: Exception

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({os.fspath(self.path)!r}, {self.exception!r})"

    # exceptions are not allowed to compare, small hack here
    def _compare_tuple(self, args) -> tuple:
        return tuple(str(arg) if isinstance(arg, Exception) else arg for arg in args)



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
"""
__version__ parameter required to be able to output to the console
"""

from pathlib import Path

VERSION_FILE = Path(__file__).parent / "VERSION"

__version__ = VERSION_FILE.read_text().strip()  # noqa: WPS410



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging

from onetl.strategy.base_strategy import BaseStrategy
from onetl.strategy.batch_hwm_strategy import BatchHWMStrategy

log = logging.getLogger(__name__)


class SnapshotStrategy(BaseStrategy):
    """Snapshot strategy for :ref:`db-reader`/:ref:`file-downloader`.

    Used for fetching all the rows/files from a source. Does not support HWM.

    .. note::

        This is a default strategy.

    For :ref:`db-reader`:
        Every snapshot run is executing the simple query which fetches all the table data:

        .. code:: sql

            SELECT id, data FROM public.mydata;

    For :ref:`file-downloader`:
        Every snapshot run is downloading all the files (from the source, or user-defined list):

        .. code:: bash

            $ hdfs dfs -ls /path

            /path/my/file1
            /path/my/file2

        .. code:: python

            assert download_result == DownloadResult(
                successful=[
                    "/path/my/file1",
                    "/path/my/file2",
                ]
            )

    Examples
    --------

    Snapshot run with :ref:`db-reader`:

    .. code:: python

        from onetl.connection import Postgres
        from onetl.db import DBReader
        from onetl.strategy import SnapshotStrategy
        from onetl.hwm import AutoDetectHWM

        from pyspark.sql import SparkSession

        maven_packages = Postgres.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        postgres = Postgres(
            host="postgres.domain.com",
            user="myuser",
            password="*****",
            database="target_database",
            spark=spark,
        )

        reader = DBReader(
            connection=postgres,
            source="public.mydata",
            columns=["id", "data"],
            hwm=DBReader.AutoDetectHWM(name="some_hwm_name", expression="id"),
        )

        writer = DBWriter(connection=hive, target="newtable")

        with SnapshotStrategy():
            df = reader.run()
            writer.run(df)

        # current run will execute following query:

        # SELECT id, data FROM public.mydata;

    Snapshot run with :ref:`file-downloader`:

    .. code:: python

        from onetl.connection import SFTP
        from onetl.file import FileDownloader
        from onetl.strategy import SnapshotStrategy

        sftp = SFTP(
            host="sftp.domain.com",
            user="user",
            password="*****",
        )

        downloader = FileDownloader(
            connection=sftp,
            source_path="/remote",
            local_path="/local",
        )

        with SnapshotStrategy():
            df = downloader.run()

        # current run will download all files from 'source_path'
    """


class SnapshotBatchStrategy(BatchHWMStrategy):
    """Snapshot batch strategy for :ref:`db-reader`.

    .. note::

        Cannot be used with :ref:`file-downloader`

    Same as :obj:`SnapshotStrategy <onetl.strategy.snapshot_strategy.SnapshotStrategy>`,
    but reads data from the source in sequential batches (1..N) like:

    .. code:: sql

        1:  SELECT id, data
            FROM public.mydata
            WHERE id >= 1000 AND id <= 1100; -- from start to start+step (INCLUDING first row)

        2:  WHERE id > 1100 AND id <= 1200; -- + step
        3:  WHERE id > 1200 AND id <= 1200; -- + step
        N:  WHERE id > 1300 AND id <= 1400; -- until stop

    This allows to use less CPU and RAM on Spark cluster than reading all the data in parallel,
    but takes proportionally more time.

    .. note::

        This strategy uses HWM column value to filter data for each batch,
        but does **NOT** save it into :ref:`HWM Store <hwm>`.
        So every run starts from the beginning, not from the previous HWM value.

    .. note::

        If you only need to reduce number of rows read by Spark from opened cursor,
        use :obj:`onetl.connection.db_connection.postgres.Postgres.ReadOptions.fetchsize` instead

    .. warning::

        Not every :ref:`DB connection <db-connections>`
        supports batch strategy. For example, Kafka connection doesn't support it.
        Make sure the connection you use is compatible with the SnapshotBatchStrategy.

    Parameters
    ----------
    step : Any

        Step size used for generating batch SQL queries like:

        .. code:: sql

            SELECT id, data
            FROM public.mydata
            WHERE id >= 1000 AND id <= 1100; -- 1000 is start value, step is 100

        .. note::

            Step defines a range of values will be fetched by each batch. This is **not**
            a number of rows, it depends on a table content and value distribution across the rows.

        .. note::

            ``step`` value will be added to the HWM, so it should have a proper type.

            For example, for ``TIMESTAMP`` column ``step`` type should be :obj:`datetime.timedelta`, not :obj:`int`

    start : Any, default: ``None``

        If passed, the value will be used for generating WHERE clauses with ``hwm.expression`` filter,
        as a start value for the first batch.

        If not set, the value is determined by a separated query:

        .. code:: sql

            SELECT MIN(id) as start
            FROM public.mydata
            WHERE id <= 1400; -- 1400 here is stop value (if set)

        .. note::

            ``start`` should be the same type as ``hwm.expression`` value,
            e.g. :obj:`datetime.datetime` for ``TIMESTAMP`` column, :obj:`datetime.date` for ``DATE``, and so on

    stop : Any, default: ``None``

        If passed, the value will be used for generating WHERE clauses with ``hwm.expression`` filter,
        as a stop value for the last batch.

        If not set, the value is determined by a separated query:

        .. code:: sql

            SELECT MAX(id) as stop
            FROM public.mydata
            WHERE id >= 1000; -- 1000 here is start value (if set)

        .. note::

            ``stop`` should be the same type as ``hwm.expression`` value,
            e.g. :obj:`datetime.datetime` for ``TIMESTAMP`` column, :obj:`datetime.date` for ``DATE``, and so on

    Examples
    --------

    SnapshotBatch run:

    .. code:: python

        from onetl.connection import Postgres, Hive
        from onetl.db import DBReader
        from onetl.strategy import SnapshotBatchStrategy
        from onetl.hwm import AutoDetectHWM

        from pyspark.sql import SparkSession

        maven_packages = Postgres.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        postgres = Postgres(
            host="postgres.domain.com",
            user="myuser",
            password="*****",
            database="target_database",
            spark=spark,
        )

        hive = Hive(cluster="rnd-dwh", spark=spark)

        reader = DBReader(
            connection=postgres,
            source="public.mydata",
            columns=["id", "data"],
            hwm=DBReader.AutoDetectHWM(name="some_hwm_name", expression="id"),
        )

        writer = DBWriter(connection=hive, target="newtable")

        with SnapshotBatchStrategy(step=100) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- get start and stop values

            SELECT MIN(id) as start, MAX(id) as stop
            FROM public.mydata;

        -- for example, start=1000 and stop=2345

        -- when each batch (1..N) will perform a query which return some part of input data

        1:  SELECT id, data
            FROM public.mydata
            WHERE id >= 1000 AND id <= 1100; -- from start to start+step (INCLUDING first row)

        2:  WHERE id > 1100 AND id <= 1200; -- + step
        3:  WHERE id > 1200 AND id <= 1300; -- + step
        N:  WHERE id > 2300 AND id <= 2345; -- until stop

    SnapshotBatch run with ``stop`` value:

    .. code:: python

        with SnapshotBatchStrategy(step=100, stop=1234) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- stop value is set, so there is no need to fetch it from DB
        -- get start value

            SELECT MIN(id) as start
            FROM public.mydata
            WHERE id <= 1234; -- until stop

        -- for example, start=1000.
        -- when each batch (1..N) will perform a query which return some part of input data

        1:  SELECT id, data
            FROM public.mydata
            WHERE id >= 1000 AND id <= 1100; -- from start to start+step (INCLUDING first row)

        2:  WHERE id >  1100 AND id <= 1200; -- + step
        3:  WHERE id >  1200 AND id <= 1300; -- + step
        N:  WHERE id >  1300 AND id <= 1234; -- until stop

    SnapshotBatch run with ``start`` value:

    .. code:: python

        with SnapshotBatchStrategy(step=100, start=500) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- start value is set, so there is no need to fetch it from DB
        -- get only stop value

            SELECT MAX(id) as stop
            FROM public.mydata
            WHERE id >= 500; -- from start

        -- for example, stop=2345.
        -- when each batch (1..N) will perform a query which return some part of input data

        1:  SELECT id, data
            FROM public.mydata
            WHERE id >= 500 AND id <=  600; -- from start to start+step (INCLUDING first row)

        2:  WHERE id >  600 AND id <=  700; -- + step
        3:  WHERE id >  700 AND id <=  800; -- + step
        ...
        N:  WHERE id > 2300 AND id <= 2345; -- until stop

    SnapshotBatch run with all options:

    .. code:: python

        with SnapshotBatchStrategy(
            start=1000,
            step=100,
            stop=2000,
        ) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- start and stop values are set, so no need to fetch boundaries from DB
        -- each batch (1..N) will perform a query which return some part of input data

        1:    SELECT id, data
            FROM public.mydata
            WHERE id >= 1000 AND id <= 1100; -- from start to start+step (INCLUDING first row)

        2:  WHERE id >  1100 AND id <= 1200; -- + step
        3:  WHERE id >  1200 AND id <= 1300; -- + step
        ...
        N:  WHERE id >  1900 AND id <= 2000; -- until stop

    ``hwm.expression`` can be a date or datetime, not only integer:

    .. code:: python

        from datetime import date, timedelta

        reader = DBReader(
            connection=postgres,
            source="public.mydata",
            columns=["business_dt", "data"],
            hwm=DBReader.AutoDetectHWM(name="some_hwm_name", expression="business_dt"),
        )

        with SnapshotBatchStrategy(
            start=date("2021-01-01"),
            step=timedelta(days=5),
            stop=date("2021-01-31"),
        ) as batches:
            for _ in batches:
                df = reader.run()
                writer.run(df)

    .. code:: sql

        -- start and stop values are set, so no need to fetch boundaries from DB
        -- each batch will perform a query which return some part of input data
        -- HWM value will casted to match column type


        1:  SELECT business_dt, data
            FROM public.mydata
            WHERE business_dt >= CAST('2020-01-01' AS DATE) -- from start to start+step (INCLUDING first row)
            AND   business_dt <= CAST('2021-01-05' AS DATE);

        2:  WHERE business_dt >  CAST('2021-01-05' AS DATE) -- + step
            AND   business_dt <= CAST('2021-01-10' AS DATE);

        3:  WHERE business_dt >  CAST('2021-01-10' AS DATE) -- + step
            AND   business_dt <= CAST('2021-01-15' AS DATE);

        ...

        N:  WHERE business_dt >  CAST('2021-01-30' AS DATE)
            AND   business_dt <= CAST('2021-01-31' AS DATE); -- until stop

    """

    def fetch_hwm(self) -> None:
        pass

    def save_hwm(self) -> None:
        pass



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar

from pydantic import Field

from onetl._util.java import try_import_java_class
from onetl._util.scala import get_default_scala_version
from onetl._util.spark import get_spark_version
from onetl._util.version import Version
from onetl.exception import MISSING_JVM_CLASS_MSG
from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import SparkSession


PROHIBITED_OPTIONS = frozenset(
    (
        # filled by onETL classes
        "path",
    ),
)


READ_OPTIONS = frozenset(
    (
        "rowTag",
        "samplingRatio",
        "excludeAttribute",
        "treatEmptyValuesAsNulls",
        "mode",
        "inferSchema",
        "columnNameOfCorruptRecord",
        "attributePrefix",
        "valueTag",
        "charset",
        "ignoreSurroundingSpaces",
        "wildcardColName",
        "rowValidationXSDPath",
        "ignoreNamespace",
        "timestampFormat",
        "dateFormat",
    ),
)

WRITE_OPTIONS = frozenset(
    (
        "rowTag",
        "rootTag",
        "declaration",
        "arrayElementName",
        "nullValue",
        "attributePrefix",
        "valueTag",
        "compression",
        "timestampFormat",
        "dateFormat",
    ),
)


log = logging.getLogger(__name__)


@support_hooks
class XML(ReadWriteFileFormat):
    """
    XML file format. |support_hooks|

    Based on `Databricks Spark XML <https://github.com/databricks/spark-xml>`_ file format.

    Supports reading/writing files with ``.xml`` extension.

    .. warning::

        Due to `bug <https://github.com/databricks/spark-xml/issues/664>`_ written files currently does not have ``.xml`` extension.

    .. versionadded:: 0.9.5

    .. dropdown:: Version compatibility

        * Spark versions: 3.2.x - 3.5.x.
        * Scala versions: 2.12 - 2.13
        * Java versions: 8 - 20

        See documentation from link above.

    .. note ::

        You can pass any option to the constructor, even if it is not mentioned in this documentation.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version. See link above.

    .. warning::

        By default, reading is done using ``mode=PERMISSIVE`` which replaces columns with wrong data type or format with ``null`` values.
        Be careful while parsing values like timestamps, they should match the ``timestampFormat`` option.
        Using ``mode=FAILFAST`` will throw an exception instead of producing ``null`` values.
        `Follow <https://github.com/databricks/spark-xml/issues/662>`_

    Examples
    --------
    Describe options how to read from/write to XML file with specific options:

    .. code:: python

        from onetl.file.format import XML
        from pyspark.sql import SparkSession

        # Create Spark session with XML package loaded
        maven_packages = XML.get_packages(spark_version="3.5.0")
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        xml = XML(row_tag="item")

    """

    name: ClassVar[str] = "xml"

    row_tag: str = Field(alias="rowTag")

    class Config:
        known_options = READ_OPTIONS | WRITE_OPTIONS
        prohibited_options = PROHIBITED_OPTIONS
        extra = "allow"

    @slot
    @classmethod
    def get_packages(  # noqa: WPS231
        cls,
        spark_version: str | Version,
        scala_version: str | Version | None = None,
        package_version: str | Version | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. |support_hooks|

        Parameters
        ----------
        spark_version : str
            Spark version in format ``major.minor.patch``.

        scala_version : str, optional
            Scala version in format ``major.minor``.

            If ``None``, ``spark_version`` is used to determine Scala version.

        package_version : str, optional
            Package version in format ``major.minor.patch``. Default is ``0.17.0``.

            See `Maven index <https://mvnrepository.com/artifact/com.databricks/spark-xml>`_
            for list of available versions.

            .. warning::

                Version ``0.13`` and below are not supported.

            .. note::

                It is not guaranteed that custom package versions are supported.
                Tests are performed only for default version.

        Examples
        --------

        .. code:: python

            from onetl.file.format import XML

            XML.get_packages(spark_version="3.5.0")
            XML.get_packages(spark_version="3.5.0", scala_version="2.12")
            XML.get_packages(
                spark_version="3.5.0",
                scala_version="2.12",
                package_version="0.17.0",
            )

        """

        if package_version:
            version = Version.parse(package_version)
            if version < (0, 14):
                raise ValueError(f"Package version must be above 0.13, got {version}")
            log.warning("Passed custom package version %r, it is not guaranteed to be supported", package_version)
        else:
            version = Version(0, 17, 0)

        spark_ver = Version.parse(spark_version)
        scala_ver = Version.parse(scala_version) if scala_version else get_default_scala_version(spark_ver)

        # Ensure compatibility with Spark and Scala versions
        if spark_ver < (3, 0):
            raise ValueError(f"Spark version must be 3.x, got {spark_ver}")

        if scala_ver < (2, 12) or scala_ver > (2, 13):
            raise ValueError(f"Scala version must be 2.12 or 2.13, got {scala_ver}")

        return [f"com.databricks:spark-xml_{scala_ver.digits(2)}:{version.digits(3)}"]

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        java_class = "com.databricks.spark.xml.XmlReader"

        try:
            try_import_java_class(spark, java_class)
        except Exception as e:
            spark_version = get_spark_version(spark)
            msg = MISSING_JVM_CLASS_MSG.format(
                java_class=java_class,
                package_source=self.__class__.__name__,
                args=f"spark_version='{spark_version}'",
            )
            if log.isEnabledFor(logging.DEBUG):
                log.debug("Missing Java class", exc_info=e, stack_info=True)
            raise ValueError(msg) from e



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from onetl.impl.base_model import BaseModel
from onetl.impl.failed_local_file import FailedLocalFile
from onetl.impl.file_exist_behavior import FileExistBehavior
from onetl.impl.frozen_model import FrozenModel
from onetl.impl.generic_options import GenericOptions
from onetl.impl.local_path import LocalPath
from onetl.impl.path_repr import path_repr
from onetl.impl.remote_directory import RemoteDirectory
from onetl.impl.remote_file import FailedRemoteFile, RemoteFile
from onetl.impl.remote_path import RemotePath
from onetl.impl.remote_path_stat import RemotePathStat



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod
from typing import TypeVar

T = TypeVar("T")


class BaseConnection(ABC):
    """
    Generic connection class
    """

    @abstractmethod
    def check(self: T) -> T:
        """Check source availability. |support_hooks|

        If not, an exception will be raised.

        Returns
        -------
        Connection itself

        Raises
        ------
        RuntimeError
            If the connection is not available

        Examples
        --------

        .. code:: python

            connection.check()
        """



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import textwrap
import warnings

from pydantic import validator

from onetl.base import BaseFileLimit, PathProtocol
from onetl.impl import FrozenModel


class FileLimit(BaseFileLimit, FrozenModel):
    """Limits the number of downloaded files.

    .. deprecated:: 0.8.0

        Use :obj:`MaxFilesCount <onetl.file.limit.max_files_count.MaxFilesCount>` instead.

    Parameters
    ----------

    count_limit : int, default = 100

        Number of downloaded files at a time.

    Examples
    --------

    Create a FileLimit object and set the amount in it:

    .. code:: python

        from onetl.core import FileLimit

        limit = FileLimit(count_limit=1500)

    If you create a :obj:`onetl.file.file_downloader.file_downloader.FileDownloader` object without
    specifying the limit option, it will download with a limit of 100 files.
    """

    count_limit: int = 100

    _counter: int = 0

    def reset(self):
        self._counter = 0
        return self

    def stops_at(self, path: PathProtocol) -> bool:
        if self.is_reached:
            return True

        if path.is_dir():
            return False

        # directories count does not matter
        self._counter += 1
        return self.is_reached

    @property
    def is_reached(self) -> bool:
        return self._counter >= self.count_limit

    @validator("count_limit")
    def _deprecated(cls, value):
        message = f"""
            Using FileLimit is deprecated since v0.8.0 and will be removed in v1.0.0.

            Please replace:
                from onetl.core import FileLimit

                limit=FileLimit(count_limit={value})

            With:
                from onetl.file.limit import MaxFilesCount

                limits=[MaxFilesCount({value})]
        """

        warnings.warn(
            textwrap.dedent(message).strip(),
            category=UserWarning,
            stacklevel=3,  # 1 is current method, 2 is BaseModel internals, 3 is user code
        )
        return value



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from contextlib import ExitStack, contextmanager
from functools import partial
from typing import TypeVar

from onetl.hooks.slot import Slot, is_slot, register_slot

logger = logging.getLogger(__name__)

Klass = TypeVar("Klass", bound=type)


def get_slots(cls: type) -> list[Slot]:
    """Returns list of methods in class decorated with :obj:`~slot`"""
    result = []
    for method_name in dir(cls):
        method = getattr(cls, method_name)
        if is_slot(method):
            result.append(method)

    return result


@contextmanager
def skip_hooks(cls: type):
    """
    Context manager (and decorator) which temporary disables hooks for all the methods of a specific class.

    Examples
    ---------

    .. tabs::

        .. code-tab:: py Context manager syntax

            @support_hooks
            class MyClass:
                @slot
                def my_method(self, arg):
                    ...


            @MyClass.my_method.hook
            def callback(self, arg):
                ...


            obj = MyClass()
            obj.my_method(1)  # will execute callback(obj, 1)

            with MyClass.skip_hooks():
                obj.my_method()  # will NOT execute callback

            # running outside the context restores previous behavior
            obj.my_method(2)  # will execute callback(obj, 2)

        .. code-tab:: py Decorator syntax

            @support_hooks
            class MyClass:
                @slot
                def my_method(self, arg):
                    ...


            @MyClass.my_method.hook
            def callback(self, arg):
                ...


            def with_hook_enabled():
                obj = MyClass()
                obj.my_method(1)


            with_hook_enabled()  # will execute callback(obj, 1)


            @MyClass.skip_hooks()
            def with_all_hooks_disabled():
                obj = MyClass()
                obj.my_method(1)


            with_all_hooks_disabled()  # will NOT execute callback function

            # running outside a decorated function restores previous behavior
            obj = MyClass()
            obj.my_method(2)  # will execute callback(obj, 2)
    """

    slots = get_slots(cls)

    with ExitStack() as stack:
        for slot in slots:
            stack.enter_context(slot.skip_hooks())
        yield


def suspend_hooks(cls: type) -> None:
    """
    Disables all hooks for all the methods of a specific class.

    Examples
    ---------

    .. code:: python

        @support_hooks
        class MyClass:
            @slot
            def my_method(self, arg): ...


        @MyClass.my_method.hook
        def callback(self, arg): ...


        obj = MyClass()
        obj.my_method(1)  # will execute callback(obj, 1)

        MyClass.suspend_hooks()

        obj.my_method(2)  # will NOT execute callback
    """

    slots = get_slots(cls)

    for slot in slots:
        slot.suspend_hooks()


def resume_hooks(cls: type) -> None:
    """
    Enables all hooks for all the methods of a specific class.

    Examples
    ---------

    .. code:: python

        @support_hooks
        class MyClass:
            @slot
            def my_method(self, arg): ...


        @MyClass.my_method.hook
        def callback(self, arg): ...


        obj = MyClass()

        MyClass.suspend_hooks()
        obj.my_method(1)  # will NOT execute callback

        MyClass.resume_hooks()

        obj.my_method(2)  # will execute callback(obj, 2)
    """

    slots = get_slots(cls)

    for slot in slots:
        slot.resume_hooks()


def support_hooks(cls: Klass) -> Klass:
    """
    Decorator which adds hooks functionality to a specific class.

    Only methods decorated with :obj:`~slot` can be used for connecting hooks.

    Adds :obj:`~skip_hooks`, :obj:`~suspend_hooks` and :obj:`~resume_hooks` to the class.

    Examples
    ---------

    .. code:: python

        from onetl.hooks.hook import support_hooks, slot


        @support_hooks
        class MyClass:
            @slot
            def my_method(self, arg): ...


        @MyClass.my_method.hook
        def callback(self, arg): ...


        MyClass().my_method()  # will execute callback function
    """

    has_slots = False
    for method_name, method in cls.__dict__.items():
        if is_slot(method):
            has_slots = True
            setattr(cls, method_name, register_slot(cls, method_name))

    if not has_slots:
        raise SyntaxError("@support_hooks can be used only with @slot decorator on some of class methods")

    cls.skip_hooks = partial(skip_hooks, cls)  # type: ignore[attr-defined]
    cls.suspend_hooks = partial(suspend_hooks, cls)  # type: ignore[attr-defined]
    cls.resume_hooks = partial(resume_hooks, cls)  # type: ignore[attr-defined]
    return cls



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import warnings
from typing import ClassVar

from onetl._util.classproperty import classproperty
from onetl._util.version import Version
from onetl.connection.db_connection.jdbc_connection import JDBCConnection
from onetl.connection.db_connection.mssql.dialect import MSSQLDialect
from onetl.hooks import slot, support_hooks
from onetl.impl import GenericOptions

# do not import PySpark here, as we allow user to use `MSSQL.get_packages()` for creating Spark session


class MSSQLExtra(GenericOptions):
    class Config:
        extra = "allow"
        prohibited_options = frozenset(("databaseName",))


@support_hooks
class MSSQL(JDBCConnection):
    """MSSQL JDBC connection. |support_hooks|

    Based on Maven package ``com.microsoft.sqlserver:mssql-jdbc:12.2.0.jre8``
    (`official MSSQL JDBC driver
    <https://docs.microsoft.com/en-us/sql/connect/jdbc/download-microsoft-jdbc-driver-for-sql-server>`_).

    .. dropdown:: Version compatibility

        * SQL Server versions: 2014 - 2022
        * Spark versions: 2.3.x - 3.5.x
        * Java versions: 8 - 20

        See `official documentation <https://learn.microsoft.com/en-us/sql/connect/jdbc/system-requirements-for-the-jdbc-driver>`_
        and `official compatibility matrix <https://learn.microsoft.com/en-us/sql/connect/jdbc/microsoft-jdbc-driver-for-sql-server-support-matrix>`_.

    .. warning::

        To use MSSQL connector you should have PySpark installed (or injected to ``sys.path``)
        BEFORE creating the connector instance.

        You can install PySpark as follows:

        .. code:: bash

            pip install onetl[spark]  # latest PySpark version

            # or
            pip install onetl pyspark=3.5.0  # pass specific PySpark version

        See :ref:`install-spark` installation instruction for more details.

    Parameters
    ----------
    host : str
        Host of MSSQL database. For example: ``test.mssql.domain.com`` or ``192.168.1.14``

    port : int, default: ``1433``
        Port of MSSQL database

    user : str
        User, which have proper access to the database. For example: ``some_user``

    password : str
        Password for database connection

    database : str
        Database in RDBMS, NOT schema.

        See `this page <https://www.educba.com/postgresql-database-vs-schema/>`_ for more details

    spark : :obj:`pyspark.sql.SparkSession`
        Spark session.

    extra : dict, default: ``None``
        Specifies one or more extra parameters by which clients can connect to the instance.

        For example: ``{"connectRetryCount": 3, "connectRetryInterval": 10}``

        See `MSSQL JDBC driver properties documentation
        <https://learn.microsoft.com/en-us/sql/connect/jdbc/setting-the-connection-properties#properties>`_
        for more details

    Examples
    --------

    MSSQL connection with plain auth:

    .. code:: python

        from onetl.connection import MSSQL
        from pyspark.sql import SparkSession

        # Create Spark session with MSSQL driver loaded
        maven_packages = MSSQL.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        # Create connection
        mssql = MSSQL(
            host="database.host.or.ip",
            user="user",
            password="*****",
            extra={
                "trustServerCertificate": "true",  # add this to avoid SSL certificate issues
            },
            spark=spark,
        )

    MSSQL connection with domain auth:

    .. code:: python

        # Create Spark session with MSSQL driver loaded
        ...

        # Create connection
        mssql = MSSQL(
            host="database.host.or.ip",
            user="user",
            password="*****",
            extra={
                "Domain": "some.domain.com",  # add here your domain
                "IntegratedSecurity": "true",
                "authenticationScheme": "NTLM",
                "trustServerCertificate": "true",  # add this to avoid SSL certificate issues
            },
            spark=spark,
        )

    MSSQL read-only connection:

    .. code:: python

        # Create Spark session with MSSQL driver loaded
        ...

        # Create connection
        mssql = MSSQL(
            host="database.host.or.ip",
            user="user",
            password="*****",
            extra={
                "ApplicationIntent": "ReadOnly",  # driver will open read-only connection, to avoid writing to the database
                "trustServerCertificate": "true",  # add this to avoid SSL certificate issues
            },
            spark=spark,
        )

    """

    database: str
    port: int = 1433
    extra: MSSQLExtra = MSSQLExtra()

    Extra = MSSQLExtra
    Dialect = MSSQLDialect

    DRIVER: ClassVar[str] = "com.microsoft.sqlserver.jdbc.SQLServerDriver"
    _CHECK_QUERY: ClassVar[str] = "SELECT 1 AS field"

    @slot
    @classmethod
    def get_packages(
        cls,
        java_version: str | Version | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. |support_hooks|

        Parameters
        ----------
        java_version : str, default ``8``
            Java major version.

        Examples
        --------

        .. code:: python

            from onetl.connection import MSSQL

            MSSQL.get_packages()
            MSSQL.get_packages(java_version="8")

        """
        if java_version is None:
            java_version = "8"

        java_ver = Version.parse(java_version)
        if java_ver.major < 8:
            raise ValueError(f"Java version must be at least 8, got {java_ver}")

        jre_ver = "8" if java_ver.major < 11 else "11"
        return [f"com.microsoft.sqlserver:mssql-jdbc:12.2.0.jre{jre_ver}"]

    @classproperty
    def package(cls) -> str:
        """Get package name to be downloaded by Spark."""
        msg = "`MSSQL.package` will be removed in 1.0.0, use `MSSQL.get_packages()` instead"
        warnings.warn(msg, UserWarning, stacklevel=3)
        return "com.microsoft.sqlserver:mssql-jdbc:12.2.0.jre8"

    @property
    def jdbc_url(self) -> str:
        prop = self.extra.dict(by_alias=True)
        prop["databaseName"] = self.database
        parameters = ";".join(f"{k}={v}" for k, v in sorted(prop.items()))

        return f"jdbc:sqlserver://{self.host}:{self.port};{parameters}"

    @property
    def instance_url(self) -> str:
        return f"{super().instance_url}/{self.database}"



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

PROHIBITED_OPTIONS = frozenset(
    (
        # These options should be passed in Spark session config, not file format options
        "spark.*",
    ),
)

READ_OPTIONS = frozenset(
    (
        "datetimeRebaseMode",
        "int96RebaseMode",
        "mergeSchema",
    ),
)

WRITE_OPTIONS = frozenset(
    (
        "compression",
        "parquet.*",
    ),
)


@support_hooks
class Parquet(ReadWriteFileFormat):
    """
    Parquet file format. |support_hooks|

    Based on `Spark Parquet Files <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html>`_ file format.

    Supports reading/writing files with ``.parquet`` extension.

    .. note ::

        You can pass any option to the constructor, even if it is not mentioned in this documentation.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version. See link above.

    Examples
    --------

    Describe options how to read from/write to Parquet file with specific options:

    .. code:: python

        parquet = Parquet(compression="snappy")

    """

    name: ClassVar[str] = "parquet"

    class Config:
        known_options = READ_OPTIONS | WRITE_OPTIONS
        prohibited_options = PROHIBITED_OPTIONS
        extra = "allow"

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        # always available
        pass



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import Optional

from pydantic import Field

from onetl.impl import GenericOptions

# options generated by JDBCMixin methods
PROHIBITED_OPTIONS = frozenset(
    (
        "user",
        "password",
        "driver",
        "url",
    ),
)


class JDBCOptions(GenericOptions):
    """Generic options, related to specific JDBC driver.

    .. note ::

        You can pass any value
        supported by underlying JDBC driver class,
        even if it is not mentioned in this documentation.
    """

    class Config:
        prohibited_options = PROHIBITED_OPTIONS
        extra = "allow"

    query_timeout: Optional[int] = Field(default=None, alias="queryTimeout")
    """The number of seconds the driver will wait for a statement to execute.
    Zero means there is no limit.

    This option depends on driver implementation,
    some drivers can check the timeout of each query instead of an entire JDBC batch.
    """

    fetchsize: Optional[int] = None
    """How many rows to fetch per round trip.

    Tuning this option can influence performance of reading.

    .. warning::

        Default value depends on driver. For example, Oracle has
        default ``fetchsize=10``.
    """



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, TypeVar

from onetl.base import BaseReadableFileFormat, BaseWritableFileFormat
from onetl.hooks import slot, support_hooks
from onetl.impl import GenericOptions

if TYPE_CHECKING:
    from pyspark.sql import DataFrameReader, DataFrameWriter

T = TypeVar("T")

PROHIBITED_OPTIONS = frozenset(
    (
        "ignoreCorruptFiles",
        "ignoreMissingFiles",
        "modifiedAfter",
        "modifiedBefore",
        "pathGlobFilter",
        "recursiveFileLookup",
    ),
)


@support_hooks
class ReadOnlyFileFormat(BaseReadableFileFormat, GenericOptions):
    name: ClassVar[str]

    class Config:
        prohibited_options = PROHIBITED_OPTIONS

    @slot
    def apply_to_reader(self, reader: DataFrameReader) -> DataFrameReader:
        options = self.dict(by_alias=True)
        return reader.format(self.name).options(**options)


@support_hooks
class WriteOnlyFileFormat(BaseWritableFileFormat, GenericOptions):
    name: ClassVar[str]

    class Config:
        prohibited_options = PROHIBITED_OPTIONS

    @slot
    def apply_to_writer(self, writer: DataFrameWriter) -> DataFrameWriter:
        options = self.dict(by_alias=True)
        return writer.format(self.name).options(**options)


class ReadWriteFileFormat(ReadOnlyFileFormat, WriteOnlyFileFormat):
    pass



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from abc import ABC, abstractmethod

from typing_extensions import Self

from onetl.base.path_protocol import PathProtocol


class BaseFileLimit(ABC):
    """
    Base file limit interface.

    Limits used by several onETL components, including :ref:`file-downloader` and :ref:`file-mover`,
    to determine if internal loop should be stopped.

    Unlike file filters, limits have internal state which can be updated or reset.
    """

    @abstractmethod
    def reset(self) -> Self:
        """
        Resets the internal limit state.

        Returns
        -------
        Returns a filter of the same type, but with non-reached state.

        It could be the same filter or a new one, this is an implementation detail.

        Examples
        --------

        .. code:: python

            assert limit.is_reached

            new_limit = limit.reset()
            assert not new_limit.is_reached
        """

    @abstractmethod
    def stops_at(self, path: PathProtocol) -> bool:
        """
        Update internal state and return current state.

        Parameters
        ----------
        path : :obj:`onetl.base.path_protocol.PathProtocol`
            Path to check

        Returns
        -------
        ``True`` if limit is reached, ``False`` otherwise.

        Examples
        --------

        .. code:: python

            from onetl.impl import LocalPath

            assert not limit.stops_at(LocalPath("/path/to/file.csv"))
            # do this multiple times
            ...

            # stopped on some input
            assert limit.stops_at(LocalPath("/path/to/another.csv"))

            # at this point, .stops_at() and .is_reached will always return True,
            # even on inputs that returned False before.
            # it will be in the same state until .reset() is called
            assert limit.stops_at(LocalPath("/path/to/file.csv"))
        """

    @property
    @abstractmethod
    def is_reached(self) -> bool:
        """
        Check if limit is reached.

        Returns
        -------
        ``True`` if limit is reached, ``False`` otherwise.

        Examples
        --------

        .. code:: python

            from onetl.impl import LocalPath

            assert not limit.is_reached

            assert not limit.stops_at(LocalPath("/path/to/file.csv"))
            assert not limit.is_reached

            assert limit.stops_at(LocalPath("/path/to/file.csv"))
            assert limit.is_reached
        """



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os
from dataclasses import dataclass, field

from onetl.base import PathStatProtocol
from onetl.impl.path_container import PathContainer
from onetl.impl.remote_path import RemotePath
from onetl.impl.remote_path_stat import RemotePathStat


@dataclass(eq=False, frozen=True)
class RemoteDirectory(PathContainer[RemotePath]):
    """
    Representation of existing remote directory
    """

    stats: PathStatProtocol = field(default_factory=RemotePathStat)

    def __post_init__(self):
        # frozen=True does not allow to change any field in __post_init__, small hack here
        object.__setattr__(self, "path", RemotePath(self.path))  # noqa: WPS609

    def is_dir(self) -> bool:
        return True

    def is_file(self) -> bool:
        return False

    def exists(self) -> bool:
        return True

    def stat(self) -> PathStatProtocol:
        return self.stats

    @property
    def parent(self) -> RemoteDirectory:
        return RemoteDirectory(self.path.parent)

    @property
    def parents(self) -> list[RemoteDirectory]:
        return [RemoteDirectory(parent) for parent in self.path.parents]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({os.fspath(self.path)!r})"



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from onetl.hooks import slot, support_hooks


@support_hooks
class KafkaSlots:
    """Kafka slots that could be implemented by third-party plugins"""

    @slot
    @staticmethod
    def normalize_cluster_name(cluster: str) -> str | None:
        """
        Normalize the given Kafka cluster name.

        This can be used to ensure that the Kafka cluster name conforms to specific naming conventions.

        Parameters
        ----------
        cluster : str
            The original Kafka cluster name.

        Returns
        -------
        str | None
            The normalized Kafka cluster name. If the hook cannot be applied, return ``None``.

        Examples
        --------

        .. code:: python

            from onetl.connection import Kafka
            from onetl.hooks import hook


            @Kafka.Slots.normalize_cluster_name.bind
            @hook
            def normalize_cluster_name(cluster: str) -> str | None:
                return cluster.lower()
        """

    @slot
    @staticmethod
    def get_known_clusters() -> set[str] | None:
        """
        Retrieve the collection of known Kafka clusters.

        This can be used to validate if the provided Kafka cluster name is recognized in the system.

        Returns
        -------
        set[str] | None
            A collection of known Kafka cluster names. If the hook cannot be applied, return ``None``.

        Examples
        --------

        .. code:: python

            from onetl.connection import Kafka
            from onetl.hooks import hook


            @Kafka.Slots.get_known_clusters.bind
            @hook
            def get_known_clusters() -> set[str] | None:
                return {"kafka-cluster", "local"}
        """

    @slot
    @staticmethod
    def normalize_address(address: str, cluster: str) -> str | None:
        """
        Normalize the given broker address for a specific Kafka cluster.

        This can be used to format the broker address according to specific rules, such as adding default ports.

        Parameters
        ----------
        address : str
            The original broker address.
        cluster : str
            The Kafka cluster name for which the address should be normalized.

        Returns
        -------
        str | None
            The normalized broker address. If the hook cannot be applied to the specific address, return ``None``.

        Examples
        --------

        .. code:: python

            from onetl.connection import Kafka
            from onetl.hooks import hook


            @Kafka.Slots.normalize_address.bind
            @hook
            def normalize_address(address: str, cluster: str) -> str | None:
                if cluster == "kafka-cluster" and ":" not in address:
                    return f"{address}:9092"
                return None
        """

    @slot
    @staticmethod
    def get_cluster_addresses(cluster: str) -> list[str] | None:
        """
        Retrieve a collection of known broker addresses for the specified Kafka cluster.

        This can be used to obtain the broker addresses dynamically.

        Parameters
        ----------
        cluster : str
            The Kafka cluster name.

        Returns
        -------
        list[str] | None
            A collection of broker addresses for the specified Kafka cluster. If the hook cannot be applied, return ``None``.

        Examples
        --------

        .. code:: python

            from onetl.connection import Kafka
            from onetl.hooks import hook


            @Kafka.Slots.get_cluster_addresses.bind
            @hook
            def get_cluster_addresses(cluster: str) -> list[str] | None:
                if cluster == "kafka_cluster":
                    return ["192.168.1.1:9092", "192.168.1.2:9092", "192.168.1.3:9092"]
                return None
        """



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from onetl._util.version import Version


def get_default_scala_version(spark_version: Version) -> Version:
    """
    Get default Scala version for specific Spark version
    """
    if spark_version.major < 3:
        return Version(2, 11)
    return Version(2, 12)



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from pydantic import Field

from onetl.file.file_result import FileResult, FileSet
from onetl.impl import FailedRemoteFile, LocalPath, RemoteFile, RemotePath


class DownloadResult(FileResult):
    """
    Representation of file download result.

    Container for file paths, divided into certain categories:

    * :obj:`successful`
    * :obj:`failed`
    * :obj:`skipped`
    * :obj:`missing`

    Examples
    --------

    Download files

    .. code:: python

        from onetl.impl import LocalPath, RemoteFile, FailedLocalFile
        from onetl.file import FileDownloader, DownloadResult

        downloader = FileDownloader(local_path="/local", ...)

        downloaded_files = downloader.run(
            [
                "/remote/file1",
                "/remote/file2",
                "/failed/file",
                "/existing/file",
                "/missing/file",
            ]
        )

        assert downloaded_files == DownloadResult(
            successful={
                LocalPath("/local/file1"),
                LocalPath("/local/file2"),
            },
            failed={FailedLocalFile("/failed/file")},
            skipped={RemoteFile("/existing/file")},
            missing={RemotePath("/missing/file")},
        )
    """

    successful: FileSet[LocalPath] = Field(default_factory=FileSet)
    "File paths (local) which were downloaded successfully"

    failed: FileSet[FailedRemoteFile] = Field(default_factory=FileSet)
    "File paths (remote) which were not downloaded because of some failure"

    skipped: FileSet[RemoteFile] = Field(default_factory=FileSet)
    "File paths (remote) which were skipped because of some reason"

    missing: FileSet[RemotePath] = Field(default_factory=FileSet)
    "File paths (remote) which are not present in the remote file system"



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
# TODO: remove in 1.0.0

import textwrap
import warnings
from importlib import import_module
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from etl_entities.hwm_store import BaseHWMStore, HWMStoreClassRegistry
    from etl_entities.hwm_store import HWMStoreStackManager as HWMStoreManager
    from etl_entities.hwm_store import (
        MemoryHWMStore,
        detect_hwm_store,
        register_hwm_store_class,
    )

    from onetl.hwm.store import (
        SparkTypeToHWM,
        YAMLHWMStore,
        default_hwm_store_class,
        register_spark_type_to_hwm_type_mapping,
    )

__all__ = [  # noqa: WPS410
    "BaseHWMStore",
    "SparkTypeToHWM",
    "register_spark_type_to_hwm_type_mapping",
    "HWMStoreClassRegistry",
    "default_hwm_store_class",
    "detect_hwm_store",
    "register_hwm_store_class",
    "HWMStoreManager",
    "MemoryHWMStore",
    "YAMLHWMStore",
]


def __getattr__(name: str):
    if name not in __all__:
        raise ImportError(f"cannot import name {name!r} from {__name__!r}")

    message = f"""
        Imports from module {__name__!r} are deprecated since v0.6.0 and will be removed in v1.0.0.
        Please use instead:

        from onetl.hwm.store import {name}
    """

    warnings.warn(textwrap.dedent(message).strip(), category=UserWarning, stacklevel=2)
    return getattr(import_module("onetl.hwm.store"), name)



# SPDX-FileCopyrightText: 2021-2024 MTS (Mobile Telesystems)
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from py4j.java_gateway import JavaGateway
    from pyspark.sql import SparkSession


def get_java_gateway(spark_session: SparkSession) -> JavaGateway:
    """
    Get py4j Java gateway object
    """
    return spark_session._sc._gateway  # noqa: WPS437  # type: ignore


def try_import_java_class(spark_session: SparkSession, name: str):
    """
    Try to import Java class by name
    """
    gateway = get_java_gateway(spark_session)
    klass = getattr(gateway.jvm, name)
    gateway.help(klass, display=False)
    return klass
